import stat
import sys
from unittest.mock import Mock

from setuptools import _shutil


def test_rmtree_readonly(monkeypatch, tmp_path):
    """Verify onerr works as expected"""

    tmp_dir = tmp_path / "with_readonly"
    tmp_dir.mkdir()
    some_file = tmp_dir.joinpath("file.txt")
    some_file.touch()
    some_file.chmod(stat.S_IREAD)

    expected_count = 1 if sys.platform.startswith("win") else 0
    chmod_fn = Mock(wraps=_shutil.attempt_chmod_verbose)
    monkeypatch.setattr(_shutil, "attempt_chmod_verbose", chmod_fn)

    _shutil.rmtree(tmp_dir)
    assert chmod_fn.call_count == expected_count
    assert not tmp_dir.is_dir()



# nif.py - functions for handling Algeria NIF numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIF, sometimes N.I.F. (Numéro d'Identification Fiscale, Algeria tax number).

The NIF was adopted by the Algerian tax authorities on 2006, replacing the NIS
number.

The NIF applies to physical persons, legal persons, legal entities,
administrative entities, local branches for foreign companies, associations,
professional organisations, etc.

The NIF consists of 15 digits, but sometimes it can be 20 digits long in order
to represent branches or secondary establishments.

More information:

* http://www.jecreemonentreprise.dz/index.php?option=com_content&view=article&id=612&Itemid=463&lang=fr
* https://www.mf.gov.dz/index.php/fr/fiscalite
* https://cnrcinfo.cnrc.dz/numero-didentification-fiscale-nif/
* https://nifenligne.mfdgi.gov.dz/
* http://nif.mfdgi.gov.dz/nif.asp

>>> validate('416001000000007')
'416001000000007'
>>> validate('408 020 000 150 039')
'408020000150039'
>>> validate('41201600000606600001')
'41201600000606600001'
>>> validate('000 216 001 808 337 13010')
'00021600180833713010'
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('X1600100000000V')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('408 020 000 150 039')
'408020000150039'
>>> format('000 216 001 808 337 13010')
'00021600180833713010'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators, removes surrounding
    whitespace.
    """
    return clean(number, ' ')


def validate(number):
    """Check if the number is a valid Algeria NIF number.

    This checks the length and formatting.
    """
    number = compact(number)
    if len(number) not in (15, 20):
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number is a valid Algeria NIF number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



import logging
import time
from threading import (
    Thread,
)
from typing import (
    Optional,
)

from opentelemetry import trace
from ._EventQueue import EventQueue
from ._Publisher import Publisher

tracer = trace.get_tracer(__name__)
logger = logging.getLogger(__name__)


class PublishQueue:
    def __init__(
        self,
        publisher: Publisher,
        event_queue: EventQueue,
    ):
        self.publisher = publisher
        self.event_queue = event_queue

        self.running = False
        self.thread: Optional[Thread] = None

    def start(self):
        self.running = True

        self.thread = Thread(
            target=self._publish_runner,
            name="PublishQueueThread",
        )
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        logger.info("Stopping PublishQueue")
        self.running = False
        self.thread.join()
        self.thread = None

    def _publish_runner(self):
        while self.running:
            try:
                self._publisher_loop()
            except Exception as ex:
                logger.exception("Error processing event", exc_info=ex)
                time.sleep(1)

    def _publisher_loop(self):
        logger.info("Executing publisher loop")
        to_publish_list = self.event_queue.get_events_that_need_to_be_published_now()
        if len(to_publish_list) == 0:
            self.event_queue.wait_for_new_events()
            return

        self._try_to_publish_events(to_publish_list=to_publish_list)

    def _try_to_publish_events(
        self,
        to_publish_list
    ):
        for to_publish in to_publish_list:
            self.publisher.publish_event(publish_detail=to_publish)



import logging
from typing import (
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

from custom_sysactions import user_must_be_logged
from sysactions import (
    action,
    show_messagebox,
)

from mw_helper import show_filtered_list_box_dialog
from api.table import get_table_service
from table_manager_api.dto import TableServiceDto
from table_manager_api.dto.exception import TableError

logger = logging.getLogger("TableActions")


@action
@user_must_be_logged
def unlink_tabs(
    pos_id,     # type: str
    tab_id,     # type: str
):
    # type: (...) -> None

    tab_picture_dto = get_table_service(pos_id=pos_id).get_service(table_id=tab_id)

    while True:
        tab_picture_dto = _unlink_tabs(
            pos_id=pos_id,
            tab_picture_dto=tab_picture_dto,
        )
        if tab_picture_dto is None:
            return


def _unlink_tabs(
    pos_id,             # type: str
    tab_picture_dto,    # type: TableServiceDto
):
    # type: (...) -> Optional[TableServiceDto]

    service_children = []
    originator_tab_id = tab_picture_dto.custom_properties.get("ORIGINATOR_TAB_ID")
    for service_child in tab_picture_dto.service_children:
        if originator_tab_id != service_child.table_id:
            service_children.append((service_child.table_name, service_child.table_id))

    if not service_children:
        return

    options = [name.encode("utf-8") for name, _ in service_children]

    selected_index = show_filtered_list_box_dialog(
        pos_id=pos_id,
        options=options,
        title="$SELECT_TAB_TO_UNLINK",
        mask="NOFILTER",
    )

    if selected_index is None:
        return

    tab_picture_dto = unlink_multiple_tabs(
        pos_id=pos_id,
        current_tab_id=tab_picture_dto.table.table_id,
        tabs_to_unlink=[service_children[selected_index][1]],
    )

    if tab_picture_dto is not None:
        from table_actions import create_table_modified_event

        create_table_modified_event(
            pos_id=pos_id,
            table_picture=eTree.tostring(tab_picture_dto.to_xml()),
        )


def unlink_multiple_tabs(
    pos_id,             # type: int
    current_tab_id,     # type: str
    tabs_to_unlink,     # type: List[str]
):
    # type: (...) -> Optional[TableServiceDto]

    try:
        table_picture_dto = get_table_service(pos_id=pos_id).unlink_tabs(
            current_tab_id=current_tab_id,
            tab_ids_to_unlink=tabs_to_unlink,
        )

        return table_picture_dto

    except TableError as ex:
        show_messagebox(
            posid=pos_id,
            message=str(ex),
        )
        logger.info("Error unlinking tabs {} to tab ID {}: {}".format(tabs_to_unlink, current_tab_id, ex))

        return None

    except (Exception,):
        show_messagebox(
            posid=pos_id,
            message="$UNLINK_TABLES_ERROR",
        )
        logger.exception("Error unlinking tabs {} to tab ID {}: ".format(tabs_to_unlink, current_tab_id))

        return None



import pyscripts

from discount_api import DiscountService
from mbcontextmessagehandler import MbContextMessageBus


def get_discount_service():
    # type: (...) -> DiscountService

    return DiscountService(
        message_bus=MbContextMessageBus(mbcontext=pyscripts.mbcontext),
    )



qwerty = "-=qwertyuiop[]asdfghjkl;'zxcvbnm,./_+QWERTYUIOP{}ASDFGHJKL:\"ZXCVBNM<>?"
dvorak = "[]',.pyfgcrl/=aoeuidhtns-;qjkxbmwvz{}\"<>PYFGCRL?+AOEUIDHTNS_:QJKXBMWVZ"


to_dvorak = str.maketrans(qwerty, dvorak)
to_qwerty = str.maketrans(dvorak, qwerty)


def translate(input, translation):
    """
    >>> translate('dvorak', to_dvorak)
    'ekrpat'
    >>> translate('qwerty', to_qwerty)
    'x,dokt'
    """
    return input.translate(translation)


def _translate_stream(stream, translation):
    """
    >>> import io
    >>> _translate_stream(io.StringIO('foo'), to_dvorak)
    urr
    """
    print(translate(stream.read(), translation))



from typing import (
    Dict,
    Optional,
)

from application.service import LoaderUpdaterService


class LoaderUpdaterInteractor(object):
    def __init__(
        self,                       # type: LoaderUpdaterInteractor
        loader_updater_service,     # type: LoaderUpdaterService
    ):
        # type: (...) -> None

        self.loader_updater_service = loader_updater_service

    def update_loaders(
        self,       # type: LoaderUpdaterInteractor
        data=None,  # type: Optional[PosUpdaterDto]
    ):
        # type: (...) -> None

        return self.loader_updater_service.update_loaders(
            restart_hv=True,
            data=data,
        )



class TableServiceException(Exception):

    def __init__(
        self,           # type: TableServiceException
        description,    # type: str
    ):
        # type: (...) -> None

        super(TableServiceException, self).__init__()

        self._description = description

    @property
    def error_description(
        self
    ):
        # type: (...) -> str

        return self._description

    def __str__(
        self
    ):
        # type: (...) -> str

        return "Error description: {0}".format(self._description)



import json

from msgbus import TK_SLCTRL_OMGR_DOSALE, FM_PARAM, TK_SYS_NAK, TK_SYS_ACK
from posot import OrderTakerException
from systools import sys_log_info


class OrderTaker(object):
    def __init__(self, pos_ot, mb_context):
        self.pos_ot = pos_ot
        self.mb_context = mb_context

    def do_sale(self, pos_id, item_id, price_list, qty=1, verify_option=True, sale_type='', after_total='0',
                apply_default_options=False, custom_properties=None):
        custom_properties = custom_properties if custom_properties else {}
        params = ('1' if self.pos_ot.blkopnotify else '0',
                  int(pos_id),
                  item_id,
                  price_list,
                  qty,
                  '1' if verify_option else '0',
                  '',
                  self.pos_ot.originatorId,
                  self.pos_ot.originatorNumber,
                  self.pos_ot.businessPeriod,
                  self.pos_ot.podType,
                  self.pos_ot.sessionId,
                  self.pos_ot.additionalInfo,
                  sale_type,
                  after_total,
                  '1' if apply_default_options else '0',
                  json.dumps(custom_properties))

        sys_log_info('POSOT do_sale: {0}'.format(params))
        req_buf = '\x00'.join((str(p) for p in params))
        msg = self.mb_context.MB_EasySendMessage('ORDERMGR%s' % pos_id, TK_SLCTRL_OMGR_DOSALE, FM_PARAM, req_buf)
        params = msg.data.split('\x00')
        if msg.token == TK_SYS_NAK:
            raise OrderTakerException(params[0], params[1])
        if msg.token == TK_SYS_ACK and len(params) > 2:
            return params[2]

        return ''



from helper import remove_accents
from pos_model import Order, XmlTenderTypeFromPOSMapper, TenderType, OrderTender

from _SatXmlPartBuilder import SatXmlPartBuilder
from models import CfeTender
from satbuilder import ContextKeys
from typing import Dict


class PgtoBuilder(SatXmlPartBuilder):
    def __init__(
        self,               # type: PgtoBuilder
        tenders_type,       # type: Dict
        layout_version,     # type: str
    ):
        # type: (...) -> None

        self.tenders_type = tenders_type

        self.cfe_tender = CfeTender(layout_version=layout_version)
        self.tenders = self._get_tenders()

    def _get_tenders(
        self,       # type: PgtoBuilder
    ):
        # type: (...) -> Dict[int, Dict]

        tenders = {}
        for tender in self.tenders_type.values():
            tenders[int(tender["tenderId"])] = tender

        return tenders

    def build_xml(
        self,           # type: PgtoBuilder
        order,          # type: Order
        context,        # type: ContextKeys
    ):
        # type: (...) -> unicode

        xml = "<pgto>"
        max_order_tenders = 10

        if len(order.tenders) <= max_order_tenders:
            for tender in order.tenders:
                tender_type = XmlTenderTypeFromPOSMapper.get(tender.type)
                if tender.type in self.tenders:
                    t_pag = self.tenders[tender.type].get("tPag")
                    if t_pag is not None:
                        tender_type = int(self.tenders[tender.type].get("tPag"))

                cad_mc = self._get_cad_mc(tender)
                xml += self._get_mp_tag(
                    tender_type=tender_type,
                    tender_amount=tender.amount,
                    cad_mc=cad_mc,
                )
        else:
            tenders_amount = {}
            for tender in order.tenders:
                cad_mc = self._get_cad_mc(tender)
                if cad_mc not in tenders_amount:
                    tenders_amount[cad_mc] = {}

                tender_type = XmlTenderTypeFromPOSMapper.get(tender.type)
                if tender_type not in tenders_amount[cad_mc]:
                    tenders_amount[cad_mc][tender_type] = 0

                tenders_amount[cad_mc][tender_type] += tender.amount

            payment_groups = 0
            for cad_mc, tenders in tenders_amount.items():
                for tender_type in tenders:
                    amount = tenders[tender_type]
                    xml += self._get_mp_tag(
                        tender_type=tender_type,
                        tender_amount=amount,
                        cad_mc=cad_mc,
                    )
                    payment_groups += 1

            if payment_groups > max_order_tenders:
                message = "$ORDER_NOT_FINALIZED_DUE_TO_MAX_SAT_PAYMENT_GROUPS_REACHED|{}"
                raise Exception(message.format(max_order_tenders))

        xml += "</pgto>"

        return xml

    def _get_mp_tag(
        self,               # type: PgtoBuilder
        tender_type,        # type: int
        tender_amount,      # type: float
        cad_mc,             # type: str
    ):
        # type: (...) -> str

        tag = "<MP><cMP>{0:>02}</cMP><vMP>{1:>0.2f}</vMP><cAdmC>{2}</cAdmC></MP>"
        cmp_tag = self.cfe_tender.get_tpag(tender_type)
        return tag.format(cmp_tag, tender_amount, cad_mc)

    def _get_cad_mc(
            self,
            tender  # type: OrderTender
    ):
        # type: (...) -> str

        if tender.type in [TenderType.credit, TenderType.debit] and tender.accreditor_sat not in ['', None]:
            return str(tender.accreditor_sat)

        others = "999"
        tender_description = str(remove_accents(tender.descr)).upper()
        if tender_description in self.tenders_type:
            tender_type_data_base = self.tenders_type[tender_description]
            cad_mc = str(tender_type_data_base.get("cAdmC", ""))
            return cad_mc if cad_mc.strip() else others

        return others



from __future__ import annotations

from setuptools import Command
from setuptools.warnings import SetuptoolsDeprecationWarning


# Would restrict to Literal["test"], but mypy doesn't support it: https://github.com/python/mypy/issues/8203
def __getattr__(name: str) -> type[_test]:
    if name == 'test':
        SetuptoolsDeprecationWarning.emit(
            "The test command is disabled and references to it are deprecated.",
            "Please remove any references to `setuptools.command.test` in all "
            "supported versions of the affected package.",
            due_date=(2024, 11, 15),
            stacklevel=2,
        )
        return _test
    raise AttributeError(name)


class _test(Command):
    """
    Stub to warn when test command is referenced or used.
    """

    description = "stub for old test command (do not use)"

    user_options = [
        ('test-module=', 'm', "Run 'test_suite' in specified module"),
        (
            'test-suite=',
            's',
            "Run single test, case or suite (e.g. 'module.test_suite')",
        ),
        ('test-runner=', 'r', "Test runner to use"),
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        raise RuntimeError("Support for the test command was removed in Setuptools 72")



# -*- coding: utf-8 -*-

import logging

from typing import (
    Any,
    Dict,
)

from old_helper import (
    remove_punctuation,
    validar_cnpj,
)
from helper import remove_accents
from pos_model import OrderTender

from models import NfeTender
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from services import TPagService
from utils import is_tef_tender

logger = logging.getLogger("FiscalWrapper")
CARD_XML = "<card><tpIntegra>{0}</tpIntegra><CNPJ>{1}</CNPJ><tBand>{2}</tBand>{3}</card>"
CARD_XML_INVALID_CNPJ = "<card><tpIntegra>2</tpIntegra></card>"


class PagBuilder(NfceXmlPartBuilder):

    def __init__(self, versao_ws, tenders_type):
        # type: (int, Dict) -> None

        self.versao_ws = versao_ws
        self.tenders_type = tenders_type

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        xml = "<pag>" if self.versao_ws not in (1, 3) else ""
        t_pag_service = TPagService(tenders_type=self.tenders_type)
        for tender in order.tenders:
            xml += "<detPag>" if self.versao_ws not in (1, 3) else "<pag>"

            t_pag = t_pag_service.get_t_pag(tender=tender)

            xml += "<tPag>{:>02}</tPag>".format(t_pag)

            if t_pag == NfeTender.outros:
                tender_desc = tender.descr or tender.type
                xml += "<xPag>{}</xPag>".format(remove_accents(tender_desc))

            total_amount = tender.amount - tender.change if self.versao_ws in (1, 3) else tender.amount
            xml += "<vPag>{:>0.2f}</vPag>".format(total_amount)

            xml += self._get_card_xml(
                tender=tender,
                t_pag=t_pag,
            )
            xml += "</detPag>" if self.versao_ws not in (1, 3) else "</pag>"

            if self.versao_ws not in (1, 3) and tender.change > 0.0:
                xml += "<vTroco>{:>0.2f}</vTroco>".format(tender.change)

        xml += "</pag>" if self.versao_ws not in (1, 3) else ""

        return xml

    def _get_card_xml(
        self,
        tender,         # type: OrderTender
        t_pag,          # type: int
    ):
        # type: (...) -> str

        if not self._is_valid_t_pag_to_create_card_xml(t_pag=int(t_pag)):
            return ""

        if is_tef_tender(tender=tender):
            if not validar_cnpj(tender.cnpj_auth):
                return CARD_XML_INVALID_CNPJ

            card_brand = self._get_card_brand(tender.t_band)
            cnpj_auth = remove_punctuation(tender.cnpj_auth.strip())
            auth_code = "<cAut>{0}</cAut>".format(tender.auth_code.strip()[-20:])
            return CARD_XML.format(1, cnpj_auth, card_brand, auth_code)

        tender_description = str(remove_accents(tender.descr)).upper()
        if tender_description not in self.tenders_type:
            return ""

        tender_type_data_base = self.tenders_type[tender_description]
        if self._tender_has_valid_cnpj(tender_type_data_base):
            t_band = tender_type_data_base.get("tBand", None)
            card_brand = self._get_card_brand(t_band)
            cnpj_auth = remove_punctuation(tender_type_data_base.get("cnpj", None))
            return CARD_XML.format(2, cnpj_auth, card_brand, "")

        return CARD_XML_INVALID_CNPJ

    @staticmethod
    def _get_card_brand(
        t_band,     # type: Union[str, int]
    ):
        card_brand = t_band or "99"
        try:
            int(card_brand)
        except ValueError:
            logger.warning("Invalid Card Brand {}".format(str(card_brand)))
            card_brand = "99"

        return card_brand

    @staticmethod
    def _tender_has_valid_cnpj(
        tender_type_data_base   # type: Dict
    ):
        # type: (...) -> bool

        card_cnpj = tender_type_data_base.get("cnpj", None)
        cnpj = validar_cnpj(card_cnpj)
        return True if cnpj is not False else False

    @staticmethod
    def _is_valid_t_pag_to_create_card_xml(
        t_pag,      # type: int
    ):
        # type: (...) -> bool

        tef_tenders = [
            NfeTender.cartao_credito,
            NfeTender.cartao_debito,
            NfeTender.pagamento_instantaneo_pix,
        ]
        return t_pag in tef_tenders



data = (
'Chey ',    # 0x00
'Thak ',    # 0x01
'Thak ',    # 0x02
'Thang ',    # 0x03
'Thayk ',    # 0x04
'Thong ',    # 0x05
'Pho ',    # 0x06
'Phok ',    # 0x07
'Hang ',    # 0x08
'Hang ',    # 0x09
'Hyen ',    # 0x0a
'Hwak ',    # 0x0b
'Wu ',    # 0x0c
'Huo ',    # 0x0d
None,    # 0x0e
None,    # 0x0f
'Zhong ',    # 0x10
None,    # 0x11
'Qing ',    # 0x12
None,    # 0x13
None,    # 0x14
'Xi ',    # 0x15
'Zhu ',    # 0x16
'Yi ',    # 0x17
'Li ',    # 0x18
'Shen ',    # 0x19
'Xiang ',    # 0x1a
'Fu ',    # 0x1b
'Jing ',    # 0x1c
'Jing ',    # 0x1d
'Yu ',    # 0x1e
None,    # 0x1f
'Hagi ',    # 0x20
None,    # 0x21
'Zhu ',    # 0x22
None,    # 0x23
None,    # 0x24
'Yi ',    # 0x25
'Du ',    # 0x26
None,    # 0x27
None,    # 0x28
None,    # 0x29
'Fan ',    # 0x2a
'Si ',    # 0x2b
'Guan ',    # 0x2c
None,    # 0x2d
None,    # 0x2e
None,    # 0x2f
None,    # 0x30
None,    # 0x31
None,    # 0x32
None,    # 0x33
None,    # 0x34
None,    # 0x35
None,    # 0x36
None,    # 0x37
None,    # 0x38
None,    # 0x39
None,    # 0x3a
None,    # 0x3b
None,    # 0x3c
None,    # 0x3d
None,    # 0x3e
None,    # 0x3f
None,    # 0x40
None,    # 0x41
None,    # 0x42
None,    # 0x43
None,    # 0x44
None,    # 0x45
None,    # 0x46
None,    # 0x47
None,    # 0x48
None,    # 0x49
None,    # 0x4a
None,    # 0x4b
None,    # 0x4c
None,    # 0x4d
None,    # 0x4e
None,    # 0x4f
None,    # 0x50
None,    # 0x51
None,    # 0x52
None,    # 0x53
None,    # 0x54
None,    # 0x55
None,    # 0x56
None,    # 0x57
None,    # 0x58
None,    # 0x59
None,    # 0x5a
None,    # 0x5b
None,    # 0x5c
None,    # 0x5d
None,    # 0x5e
None,    # 0x5f
None,    # 0x60
None,    # 0x61
None,    # 0x62
None,    # 0x63
None,    # 0x64
None,    # 0x65
None,    # 0x66
None,    # 0x67
None,    # 0x68
None,    # 0x69
None,    # 0x6a
None,    # 0x6b
None,    # 0x6c
None,    # 0x6d
None,    # 0x6e
None,    # 0x6f
None,    # 0x70
None,    # 0x71
None,    # 0x72
None,    # 0x73
None,    # 0x74
None,    # 0x75
None,    # 0x76
None,    # 0x77
None,    # 0x78
None,    # 0x79
None,    # 0x7a
None,    # 0x7b
None,    # 0x7c
None,    # 0x7d
None,    # 0x7e
None,    # 0x7f
None,    # 0x80
None,    # 0x81
None,    # 0x82
None,    # 0x83
None,    # 0x84
None,    # 0x85
None,    # 0x86
None,    # 0x87
None,    # 0x88
None,    # 0x89
None,    # 0x8a
None,    # 0x8b
None,    # 0x8c
None,    # 0x8d
None,    # 0x8e
None,    # 0x8f
None,    # 0x90
None,    # 0x91
None,    # 0x92
None,    # 0x93
None,    # 0x94
None,    # 0x95
None,    # 0x96
None,    # 0x97
None,    # 0x98
None,    # 0x99
None,    # 0x9a
None,    # 0x9b
None,    # 0x9c
None,    # 0x9d
None,    # 0x9e
None,    # 0x9f
None,    # 0xa0
None,    # 0xa1
None,    # 0xa2
None,    # 0xa3
None,    # 0xa4
None,    # 0xa5
None,    # 0xa6
None,    # 0xa7
None,    # 0xa8
None,    # 0xa9
None,    # 0xaa
None,    # 0xab
None,    # 0xac
None,    # 0xad
None,    # 0xae
None,    # 0xaf
None,    # 0xb0
None,    # 0xb1
None,    # 0xb2
None,    # 0xb3
None,    # 0xb4
None,    # 0xb5
None,    # 0xb6
None,    # 0xb7
None,    # 0xb8
None,    # 0xb9
None,    # 0xba
None,    # 0xbb
None,    # 0xbc
None,    # 0xbd
None,    # 0xbe
None,    # 0xbf
None,    # 0xc0
None,    # 0xc1
None,    # 0xc2
None,    # 0xc3
None,    # 0xc4
None,    # 0xc5
None,    # 0xc6
None,    # 0xc7
None,    # 0xc8
None,    # 0xc9
None,    # 0xca
None,    # 0xcb
None,    # 0xcc
None,    # 0xcd
None,    # 0xce
None,    # 0xcf
None,    # 0xd0
None,    # 0xd1
None,    # 0xd2
None,    # 0xd3
None,    # 0xd4
None,    # 0xd5
None,    # 0xd6
None,    # 0xd7
None,    # 0xd8
None,    # 0xd9
None,    # 0xda
None,    # 0xdb
None,    # 0xdc
None,    # 0xdd
None,    # 0xde
None,    # 0xdf
None,    # 0xe0
None,    # 0xe1
None,    # 0xe2
None,    # 0xe3
None,    # 0xe4
None,    # 0xe5
None,    # 0xe6
None,    # 0xe7
None,    # 0xe8
None,    # 0xe9
None,    # 0xea
None,    # 0xeb
None,    # 0xec
None,    # 0xed
None,    # 0xee
None,    # 0xef
None,    # 0xf0
None,    # 0xf1
None,    # 0xf2
None,    # 0xf3
None,    # 0xf4
None,    # 0xf5
None,    # 0xf6
None,    # 0xf7
None,    # 0xf8
None,    # 0xf9
None,    # 0xfa
None,    # 0xfb
None,    # 0xfc
None,    # 0xfd
None,    # 0xfe
)



# -*- coding: utf-8 -*-

from sysactions import action, show_messagebox
from msgbus import TK_SYS_NAK

from api.pos_updater import get_pos_updater_service
from .update_util import show_nak_message
from ..util.get_menu_user_logged import get_menu_manager_authenticated_user


@action
def update_binaries(pos_id):

    resp = show_messagebox(pos_id, "$UPDATE_BINARIES_MESSAGE_QUESTION", buttons="$OK|$CANCEL")
    if resp == 1:
        return

    user_id = get_menu_manager_authenticated_user(pos_id=int(pos_id))
    updater_service = get_pos_updater_service(pos_id=pos_id)
    reply = updater_service.binaries_update(
        user_id=user_id,
        force_update=True,
    )
    if reply.token == TK_SYS_NAK:
        show_nak_message(
            pos_id=pos_id,
            update_key="$UPDATE_BINARIES_",
            message=reply.data,
        )

    show_messagebox(pos_id, "$UPDATE_BINARIES_MESSAGE", buttons="OK")



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Internal utilities for gRPC Python."""

import collections
import logging
import threading
import time
from typing import Callable, Dict, Optional, Sequence

import grpc  # pytype: disable=pyi-error
from grpc import _common  # pytype: disable=pyi-error
from grpc._typing import DoneCallbackType

_LOGGER = logging.getLogger(__name__)

_DONE_CALLBACK_EXCEPTION_LOG_MESSAGE = (
    'Exception calling connectivity future "done" callback!'
)


class RpcMethodHandler(
    collections.namedtuple(
        "_RpcMethodHandler",
        (
            "request_streaming",
            "response_streaming",
            "request_deserializer",
            "response_serializer",
            "unary_unary",
            "unary_stream",
            "stream_unary",
            "stream_stream",
        ),
    ),
    grpc.RpcMethodHandler,
):
    pass


class DictionaryGenericHandler(grpc.ServiceRpcHandler):
    _name: str
    _method_handlers: Dict[str, grpc.RpcMethodHandler]

    def __init__(
        self, service: str, method_handlers: Dict[str, grpc.RpcMethodHandler]
    ):
        self._name = service
        self._method_handlers = {
            _common.fully_qualified_method(service, method): method_handler
            for method, method_handler in method_handlers.items()
        }

    def service_name(self) -> str:
        return self._name

    def service(
        self, handler_call_details: grpc.HandlerCallDetails
    ) -> Optional[grpc.RpcMethodHandler]:
        details_method = handler_call_details.method
        return self._method_handlers.get(
            details_method
        )  # pytype: disable=attribute-error


class _ChannelReadyFuture(grpc.Future):
    _condition: threading.Condition
    _channel: grpc.Channel
    _matured: bool
    _cancelled: bool
    _done_callbacks: Sequence[Callable]

    def __init__(self, channel: grpc.Channel):
        self._condition = threading.Condition()
        self._channel = channel

        self._matured = False
        self._cancelled = False
        self._done_callbacks = []

    def _block(self, timeout: Optional[float]) -> None:
        until = None if timeout is None else time.time() + timeout
        with self._condition:
            while True:
                if self._cancelled:
                    raise grpc.FutureCancelledError()
                elif self._matured:
                    return
                else:
                    if until is None:
                        self._condition.wait()
                    else:
                        remaining = until - time.time()
                        if remaining < 0:
                            raise grpc.FutureTimeoutError()
                        else:
                            self._condition.wait(timeout=remaining)

    def _update(self, connectivity: Optional[grpc.ChannelConnectivity]) -> None:
        with self._condition:
            if (
                not self._cancelled
                and connectivity is grpc.ChannelConnectivity.READY
            ):
                self._matured = True
                self._channel.unsubscribe(self._update)
                self._condition.notify_all()
                done_callbacks = tuple(self._done_callbacks)
                self._done_callbacks = None
            else:
                return

        for done_callback in done_callbacks:
            try:
                done_callback(self)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception(_DONE_CALLBACK_EXCEPTION_LOG_MESSAGE)

    def cancel(self) -> bool:
        with self._condition:
            if not self._matured:
                self._cancelled = True
                self._channel.unsubscribe(self._update)
                self._condition.notify_all()
                done_callbacks = tuple(self._done_callbacks)
                self._done_callbacks = None
            else:
                return False

        for done_callback in done_callbacks:
            try:
                done_callback(self)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception(_DONE_CALLBACK_EXCEPTION_LOG_MESSAGE)

        return True

    def cancelled(self) -> bool:
        with self._condition:
            return self._cancelled

    def running(self) -> bool:
        with self._condition:
            return not self._cancelled and not self._matured

    def done(self) -> bool:
        with self._condition:
            return self._cancelled or self._matured

    def result(self, timeout: Optional[float] = None) -> None:
        self._block(timeout)

    def exception(self, timeout: Optional[float] = None) -> None:
        self._block(timeout)

    def traceback(self, timeout: Optional[float] = None) -> None:
        self._block(timeout)

    def add_done_callback(self, fn: DoneCallbackType):
        with self._condition:
            if not self._cancelled and not self._matured:
                self._done_callbacks.append(fn)
                return

        fn(self)

    def start(self):
        with self._condition:
            self._channel.subscribe(self._update, try_to_connect=True)

    def __del__(self):
        with self._condition:
            if not self._cancelled and not self._matured:
                self._channel.unsubscribe(self._update)


def channel_ready_future(channel: grpc.Channel) -> _ChannelReadyFuture:
    ready_future = _ChannelReadyFuture(channel)
    ready_future.start()
    return ready_future



from ._AutoOrderCommand import AutoOrderCommand
from ._BumpBar import BumpBar
from ._BumpBarCommand import BumpBarCommand
from ._Kds import Kds
from ._KdsModel import KdsModel
from ._KdsView import KdsView
from ._Modifier import Modifier
from ._PickupArea import PickupArea
from ._PickupConfig import PickupConfig
from ._ZoomLevel import ZoomLevel



from logging import Logger

from production.model import OrderChanger
from production.view import ProductionView
from ._OrderChangerProductionBox import OrderChangerProductionBox


class StampItemsViewBox(OrderChangerProductionBox):
    def __init__(self, name, order_changer, view, sons, logger=None):
        # type: (str, OrderChanger, ProductionView, str, Logger) -> None
        super(StampItemsViewBox, self).__init__(name, sons, logger)
        self.order_changer = order_changer
        self.view = view  # type: ProductionView

    def change_order(self, order):
        def stamp_item(item_to_change):
            if order.prod_state != "INVALID" and item_to_change.tags.get('kds_view') != self.view.view_name:
                return item_to_change.add_view(self.view.view_name)

        self.order_changer.change_items(order, stamp_item)

        return order



from ._DiscountAppliedInteractor import DiscountAppliedInteractor
from ._GetConfiguredAlertsInteractor import GetConfiguredAlertsInteractor
from ._OrderCanceledInteractor import OrderCanceledInteractor
from ._POSAlertNotificationInteractor import POSAlertNotificationInteractor
from ._BOHChangeNotificationConfigInteractor import BOHChangeNotificationConfigInteractor
from ._BOHOrderPictureInteractor import BOHOrderPictureInteractor
from ._UpdateAlertsInteractor import UpdateAlertsInteractor



import contextlib
import pathlib
from pathlib import Path
import re
import time
from typing import Union
from unittest import mock


def flatten_result(result):
    return re.sub(r"[\s\r\n]+", " ", result).strip()


def result_lines(result):
    return [
        x.strip()
        for x in re.split(r"\r?\n", re.sub(r" +", " ", result))
        if x.strip() != ""
    ]


def result_raw_lines(result):
    return [x for x in re.split(r"\r?\n", result) if x.strip() != ""]


def make_path(
    filespec: Union[Path, str],
    make_absolute: bool = True,
    check_exists: bool = False,
) -> Path:
    path = Path(filespec)
    if make_absolute:
        path = path.resolve(strict=check_exists)
    if check_exists and (not path.exists()):
        raise FileNotFoundError(f"No file or directory at {filespec}")
    return path


def _unlink_path(path, missing_ok=False):
    # Replicate 3.8+ functionality in 3.7
    cm = contextlib.nullcontext()
    if missing_ok:
        cm = contextlib.suppress(FileNotFoundError)

    with cm:
        path.unlink()


def replace_file_with_dir(pathspec):
    path = pathlib.Path(pathspec)
    _unlink_path(path, missing_ok=True)
    path.mkdir(exist_ok=True)
    return path


def file_with_template_code(filespec):
    with open(filespec, "w") as f:
        f.write(
            """
i am an artificial template just for you
"""
        )
    return filespec


@contextlib.contextmanager
def rewind_compile_time(hours=1):
    rewound = time.time() - (hours * 3_600)
    with mock.patch("mako.codegen.time") as codegen_time:
        codegen_time.time.return_value = rewound
        yield



class ServiceChild(object):

    def __init__(
        self,           # type: ServiceChild
        service_id,     # type: int
        table_id,       # type: str
        status,         # type: int
        table_name,     # type: str
    ):
        # type: (...) -> None

        self.service_id = service_id
        self.table_id = table_id
        self.status = status
        self.table_name = table_name



# -*- coding: utf-8 -*-

from datetime import datetime

from nfce import NfceRequestBuilder as RequestBuilder


class NfceCancelerParameters(object):
    ModeloNfce = "65"
    AmbienteProducao = 1
    AmbienteHomologacao = 2
    InfEventoTemplate = "<evento versao=\"1.00\"><infEvento Id='ID110111{chave}01'><cOrgao>{cUF}</cOrgao><tpAmb>{tpAmb}</tpAmb><CNPJ>{cnpj}</CNPJ><chNFe>{chave}</chNFe>" \
                      "<dhEvento>{dh}</dhEvento><tpEvento>110111</tpEvento><nSeqEvento>1</nSeqEvento>" \
                      "<verEvento>1.00</verEvento><detEvento versao=\"1.00\"><descEvento>Cancelamento</descEvento>" \
                      "<nProt>{nprot}</nProt><xJust>{justificativa}</xJust></detEvento></infEvento></evento>"

    def __init__(self, tp_amb, c_uf, cnpj):
        # type: (int, int, str) -> None
        self.tp_amb = str(tp_amb)
        self.c_uf = str(c_uf)
        self.cnpj = cnpj

    def fill_parameters(self, chave, nprot, justificativa):
        # type: (str, str, str) -> str

        dh = RequestBuilder.formata_data(datetime.now())
        return NfceCancelerParameters.InfEventoTemplate \
            .format(chave=chave, tpAmb=self.tp_amb, cUF=self.c_uf, cnpj=self.cnpj, justificativa=justificativa, dh=dh, nprot=nprot)



from xml.etree import cElementTree as eTree

import pyscripts
from mbcontextmessagehandler import MbContextMessageBus
from pos_api import PosService
from typing import Optional

from thread_local import (
    get_thread_var,
    set_thread_var,
    get_current_model,
)


def get_pos_service(
    pos_id,     # type: int
    model=None  # type: Optional[eTree.Element]
):
    # type: (...) -> PosService

    if get_thread_var("pos_service"):
        return get_thread_var("pos_service")

    if model is None:
        model = get_current_model()

    pos_service = PosService(
        message_bus=MbContextMessageBus(mbcontext=pyscripts.mbcontext),
        pos_id=pos_id,
        model=model,
    )

    set_thread_var("pos_service", pos_service)
    return pos_service



from ._AddTagLinesCommand import AddTagLinesCommand
from ._ChangeProdStateCommand import ChangeProdStateCommand
from ._Command import Command
from ._CommandType import CommandType
from ._DevicesUpdatedCommand import DevicesUpdatedCommand
from ._OrderModifiedCommand import OrderModifiedCommand
from ._ProductionCommand import ProductionCommand
from ._ProductionCommandProcessor import ProductionCommandProcessor
from ._PublishOrderCommand import PublishOrderCommand
from ._PurgeOrdersCommand import PurgeOrdersCommand
from ._ToggleTagLinesCommand import ToggleTagLinesCommand
from ._UndoProductionCommand import UndoProductionCommand



class ReportNotImplemented(Exception):

    def __init__(
        self,
        report_name: str,
    ) -> None:
        super(ReportNotImplemented, self).__init__(f"Report requested was not implemented: {report_name}")



from application.domain import TableService  # noqa
from application.domain.ui import ScreenChanger, ScreenIdEnum  # noqa
from application.interactor.service import TableReleaser  # noqa
from typing import (
    Dict,
    Optional,
    Union,
)


class StartServiceInteractor(object):
    def __init__(self, screen_changer, table_service, table_releaser):
        # type: (ScreenChanger, TableService, TableReleaser) -> None
        self.screen_changer = screen_changer
        self.table_service = table_service
        self.table_releaser = table_releaser

    def execute(
        self,                       # type: StartServiceInteractor
        pos_id,                     # type: Union[int, str]
        table_id="",                # type: str
        seats="",                   # type: str
        name="",                    # type: str
        mode="",                    # type: str
        source_table_id="",         # type: str
        custom_properties=None,     # type: Optional[Dict[str, str]]
    ):
        # type: (...) -> str

        self.table_releaser.release_all(pos_id)
        ret = self.table_service.start_service(
            pos_id=pos_id,
            table_id=table_id,
            seats=seats,
            name=name,
            mode=mode,
            source_table_id=source_table_id,
            custom_properties=custom_properties,
        )
        return ret



# -*- coding: utf-8 -*-
"""
A test helper for seeing what happens when slp_switch()
fails.
"""
# pragma: no cover

import greenlet


print('fail_slp_switch is running', flush=True)

runs = []
def func():
    runs.append(1)
    greenlet.getcurrent().parent.switch()
    runs.append(2)
    greenlet.getcurrent().parent.switch()
    runs.append(3)

g = greenlet._greenlet.UnswitchableGreenlet(func)
g.switch()
assert runs == [1]
g.switch()
assert runs == [1, 2]
g.force_slp_switch_error = True

# This should crash.
g.switch()



#
# The Python Imaging Library
# $Id$
#
# Simple PostScript graphics interface
#
# History:
# 1996-04-20 fl   Created
# 1999-01-10 fl   Added gsave/grestore to image method
# 2005-05-04 fl   Fixed floating point issue in image (from Eric Etheridge)
#
# Copyright (c) 1997-2005 by Secret Labs AB.  All rights reserved.
# Copyright (c) 1996 by Fredrik Lundh.
#
# See the README file for information on usage and redistribution.
#
from __future__ import annotations

import sys

from . import EpsImagePlugin

##
# Simple PostScript graphics interface.


class PSDraw:
    """
    Sets up printing to the given file. If ``fp`` is omitted,
    ``sys.stdout.buffer`` or ``sys.stdout`` is assumed.
    """

    def __init__(self, fp=None):
        if not fp:
            try:
                fp = sys.stdout.buffer
            except AttributeError:
                fp = sys.stdout
        self.fp = fp

    def begin_document(self, id=None):
        """Set up printing of a document. (Write PostScript DSC header.)"""
        # FIXME: incomplete
        self.fp.write(
            b"%!PS-Adobe-3.0\n"
            b"save\n"
            b"/showpage { } def\n"
            b"%%EndComments\n"
            b"%%BeginDocument\n"
        )
        # self.fp.write(ERROR_PS)  # debugging!
        self.fp.write(EDROFF_PS)
        self.fp.write(VDI_PS)
        self.fp.write(b"%%EndProlog\n")
        self.isofont = {}

    def end_document(self):
        """Ends printing. (Write PostScript DSC footer.)"""
        self.fp.write(b"%%EndDocument\nrestore showpage\n%%End\n")
        if hasattr(self.fp, "flush"):
            self.fp.flush()

    def setfont(self, font, size):
        """
        Selects which font to use.

        :param font: A PostScript font name
        :param size: Size in points.
        """
        font = bytes(font, "UTF-8")
        if font not in self.isofont:
            # reencode font
            self.fp.write(b"/PSDraw-%s ISOLatin1Encoding /%s E\n" % (font, font))
            self.isofont[font] = 1
        # rough
        self.fp.write(b"/F0 %d /PSDraw-%s F\n" % (size, font))

    def line(self, xy0, xy1):
        """
        Draws a line between the two points. Coordinates are given in
        PostScript point coordinates (72 points per inch, (0, 0) is the lower
        left corner of the page).
        """
        self.fp.write(b"%d %d %d %d Vl\n" % (*xy0, *xy1))

    def rectangle(self, box):
        """
        Draws a rectangle.

        :param box: A tuple of four integers, specifying left, bottom, width and
           height.
        """
        self.fp.write(b"%d %d M 0 %d %d Vr\n" % box)

    def text(self, xy, text):
        """
        Draws text at the given position. You must use
        :py:meth:`~PIL.PSDraw.PSDraw.setfont` before calling this method.
        """
        text = bytes(text, "UTF-8")
        text = b"\\(".join(text.split(b"("))
        text = b"\\)".join(text.split(b")"))
        xy += (text,)
        self.fp.write(b"%d %d M (%s) S\n" % xy)

    def image(self, box, im, dpi=None):
        """Draw a PIL image, centered in the given box."""
        # default resolution depends on mode
        if not dpi:
            if im.mode == "1":
                dpi = 200  # fax
            else:
                dpi = 100  # grayscale
        # image size (on paper)
        x = im.size[0] * 72 / dpi
        y = im.size[1] * 72 / dpi
        # max allowed size
        xmax = float(box[2] - box[0])
        ymax = float(box[3] - box[1])
        if x > xmax:
            y = y * xmax / x
            x = xmax
        if y > ymax:
            x = x * ymax / y
            y = ymax
        dx = (xmax - x) / 2 + box[0]
        dy = (ymax - y) / 2 + box[1]
        self.fp.write(b"gsave\n%f %f translate\n" % (dx, dy))
        if (x, y) != im.size:
            # EpsImagePlugin._save prints the image at (0,0,xsize,ysize)
            sx = x / im.size[0]
            sy = y / im.size[1]
            self.fp.write(b"%f %f scale\n" % (sx, sy))
        EpsImagePlugin._save(im, self.fp, None, 0)
        self.fp.write(b"\ngrestore\n")


# --------------------------------------------------------------------
# PostScript driver

#
# EDROFF.PS -- PostScript driver for Edroff 2
#
# History:
# 94-01-25 fl: created (edroff 2.04)
#
# Copyright (c) Fredrik Lundh 1994.
#


EDROFF_PS = b"""\
/S { show } bind def
/P { moveto show } bind def
/M { moveto } bind def
/X { 0 rmoveto } bind def
/Y { 0 exch rmoveto } bind def
/E {    findfont
        dup maxlength dict begin
        {
                1 index /FID ne { def } { pop pop } ifelse
        } forall
        /Encoding exch def
        dup /FontName exch def
        currentdict end definefont pop
} bind def
/F {    findfont exch scalefont dup setfont
        [ exch /setfont cvx ] cvx bind def
} bind def
"""

#
# VDI.PS -- PostScript driver for VDI meta commands
#
# History:
# 94-01-25 fl: created (edroff 2.04)
#
# Copyright (c) Fredrik Lundh 1994.
#

VDI_PS = b"""\
/Vm { moveto } bind def
/Va { newpath arcn stroke } bind def
/Vl { moveto lineto stroke } bind def
/Vc { newpath 0 360 arc closepath } bind def
/Vr {   exch dup 0 rlineto
        exch dup 0 exch rlineto
        exch neg 0 rlineto
        0 exch neg rlineto
        setgray fill } bind def
/Tm matrix def
/Ve {   Tm currentmatrix pop
        translate scale newpath 0 0 .5 0 360 arc closepath
        Tm setmatrix
} bind def
/Vf { currentgray exch setgray fill setgray } bind def
"""

#
# ERROR.PS -- Error handler
#
# History:
# 89-11-21 fl: created (pslist 1.10)
#

ERROR_PS = b"""\
/landscape false def
/errorBUF 200 string def
/errorNL { currentpoint 10 sub exch pop 72 exch moveto } def
errordict begin /handleerror {
    initmatrix /Courier findfont 10 scalefont setfont
    newpath 72 720 moveto $error begin /newerror false def
    (PostScript Error) show errorNL errorNL
    (Error: ) show
        /errorname load errorBUF cvs show errorNL errorNL
    (Command: ) show
        /command load dup type /stringtype ne { errorBUF cvs } if show
        errorNL errorNL
    (VMstatus: ) show
        vmstatus errorBUF cvs show ( bytes available, ) show
        errorBUF cvs show ( bytes used at level ) show
        errorBUF cvs show errorNL errorNL
    (Operand stargck: ) show errorNL /ostargck load {
        dup type /stringtype ne { errorBUF cvs } if 72 0 rmoveto show errorNL
    } forall errorNL
    (Execution stargck: ) show errorNL /estargck load {
        dup type /stringtype ne { errorBUF cvs } if 72 0 rmoveto show errorNL
    } forall
    end showpage
} def end
"""



"""Stubs for more_itertools.more"""

from __future__ import annotations

from types import TracebackType
from typing import (
    Any,
    Callable,
    Container,
    ContextManager,
    Generic,
    Hashable,
    Mapping,
    Iterable,
    Iterator,
    Mapping,
    overload,
    Reversible,
    Sequence,
    Sized,
    Type,
    TypeVar,
    type_check_only,
)
from typing_extensions import Protocol

# Type and type variable definitions
_T = TypeVar('_T')
_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')
_U = TypeVar('_U')
_V = TypeVar('_V')
_W = TypeVar('_W')
_T_co = TypeVar('_T_co', covariant=True)
_GenFn = TypeVar('_GenFn', bound=Callable[..., Iterator[Any]])
_Raisable = BaseException | Type[BaseException]

@type_check_only
class _SizedIterable(Protocol[_T_co], Sized, Iterable[_T_co]): ...

@type_check_only
class _SizedReversible(Protocol[_T_co], Sized, Reversible[_T_co]): ...

@type_check_only
class _SupportsSlicing(Protocol[_T_co]):
    def __getitem__(self, __k: slice) -> _T_co: ...

def chunked(
    iterable: Iterable[_T], n: int | None, strict: bool = ...
) -> Iterator[list[_T]]: ...
@overload
def first(iterable: Iterable[_T]) -> _T: ...
@overload
def first(iterable: Iterable[_T], default: _U) -> _T | _U: ...
@overload
def last(iterable: Iterable[_T]) -> _T: ...
@overload
def last(iterable: Iterable[_T], default: _U) -> _T | _U: ...
@overload
def nth_or_last(iterable: Iterable[_T], n: int) -> _T: ...
@overload
def nth_or_last(iterable: Iterable[_T], n: int, default: _U) -> _T | _U: ...

class peekable(Generic[_T], Iterator[_T]):
    def __init__(self, iterable: Iterable[_T]) -> None: ...
    def __iter__(self) -> peekable[_T]: ...
    def __bool__(self) -> bool: ...
    @overload
    def peek(self) -> _T: ...
    @overload
    def peek(self, default: _U) -> _T | _U: ...
    def prepend(self, *items: _T) -> None: ...
    def __next__(self) -> _T: ...
    @overload
    def __getitem__(self, index: int) -> _T: ...
    @overload
    def __getitem__(self, index: slice) -> list[_T]: ...

def consumer(func: _GenFn) -> _GenFn: ...
def ilen(iterable: Iterable[_T]) -> int: ...
def iterate(func: Callable[[_T], _T], start: _T) -> Iterator[_T]: ...
def with_iter(
    context_manager: ContextManager[Iterable[_T]],
) -> Iterator[_T]: ...
def one(
    iterable: Iterable[_T],
    too_short: _Raisable | None = ...,
    too_long: _Raisable | None = ...,
) -> _T: ...
def raise_(exception: _Raisable, *args: Any) -> None: ...
def strictly_n(
    iterable: Iterable[_T],
    n: int,
    too_short: _GenFn | None = ...,
    too_long: _GenFn | None = ...,
) -> list[_T]: ...
def distinct_permutations(
    iterable: Iterable[_T], r: int | None = ...
) -> Iterator[tuple[_T, ...]]: ...
def intersperse(
    e: _U, iterable: Iterable[_T], n: int = ...
) -> Iterator[_T | _U]: ...
def unique_to_each(*iterables: Iterable[_T]) -> list[list[_T]]: ...
@overload
def windowed(
    seq: Iterable[_T], n: int, *, step: int = ...
) -> Iterator[tuple[_T | None, ...]]: ...
@overload
def windowed(
    seq: Iterable[_T], n: int, fillvalue: _U, step: int = ...
) -> Iterator[tuple[_T | _U, ...]]: ...
def substrings(iterable: Iterable[_T]) -> Iterator[tuple[_T, ...]]: ...
def substrings_indexes(
    seq: Sequence[_T], reverse: bool = ...
) -> Iterator[tuple[Sequence[_T], int, int]]: ...

class bucket(Generic[_T, _U], Container[_U]):
    def __init__(
        self,
        iterable: Iterable[_T],
        key: Callable[[_T], _U],
        validator: Callable[[_U], object] | None = ...,
    ) -> None: ...
    def __contains__(self, value: object) -> bool: ...
    def __iter__(self) -> Iterator[_U]: ...
    def __getitem__(self, value: object) -> Iterator[_T]: ...

def spy(
    iterable: Iterable[_T], n: int = ...
) -> tuple[list[_T], Iterator[_T]]: ...
def interleave(*iterables: Iterable[_T]) -> Iterator[_T]: ...
def interleave_longest(*iterables: Iterable[_T]) -> Iterator[_T]: ...
def interleave_evenly(
    iterables: list[Iterable[_T]], lengths: list[int] | None = ...
) -> Iterator[_T]: ...
def collapse(
    iterable: Iterable[Any],
    base_type: type | None = ...,
    levels: int | None = ...,
) -> Iterator[Any]: ...
@overload
def side_effect(
    func: Callable[[_T], object],
    iterable: Iterable[_T],
    chunk_size: None = ...,
    before: Callable[[], object] | None = ...,
    after: Callable[[], object] | None = ...,
) -> Iterator[_T]: ...
@overload
def side_effect(
    func: Callable[[list[_T]], object],
    iterable: Iterable[_T],
    chunk_size: int,
    before: Callable[[], object] | None = ...,
    after: Callable[[], object] | None = ...,
) -> Iterator[_T]: ...
def sliced(
    seq: _SupportsSlicing[_T], n: int, strict: bool = ...
) -> Iterator[_T]: ...
def split_at(
    iterable: Iterable[_T],
    pred: Callable[[_T], object],
    maxsplit: int = ...,
    keep_separator: bool = ...,
) -> Iterator[list[_T]]: ...
def split_before(
    iterable: Iterable[_T], pred: Callable[[_T], object], maxsplit: int = ...
) -> Iterator[list[_T]]: ...
def split_after(
    iterable: Iterable[_T], pred: Callable[[_T], object], maxsplit: int = ...
) -> Iterator[list[_T]]: ...
def split_when(
    iterable: Iterable[_T],
    pred: Callable[[_T, _T], object],
    maxsplit: int = ...,
) -> Iterator[list[_T]]: ...
def split_into(
    iterable: Iterable[_T], sizes: Iterable[int | None]
) -> Iterator[list[_T]]: ...
@overload
def padded(
    iterable: Iterable[_T],
    *,
    n: int | None = ...,
    next_multiple: bool = ...,
) -> Iterator[_T | None]: ...
@overload
def padded(
    iterable: Iterable[_T],
    fillvalue: _U,
    n: int | None = ...,
    next_multiple: bool = ...,
) -> Iterator[_T | _U]: ...
@overload
def repeat_last(iterable: Iterable[_T]) -> Iterator[_T]: ...
@overload
def repeat_last(iterable: Iterable[_T], default: _U) -> Iterator[_T | _U]: ...
def distribute(n: int, iterable: Iterable[_T]) -> list[Iterator[_T]]: ...
@overload
def stagger(
    iterable: Iterable[_T],
    offsets: _SizedIterable[int] = ...,
    longest: bool = ...,
) -> Iterator[tuple[_T | None, ...]]: ...
@overload
def stagger(
    iterable: Iterable[_T],
    offsets: _SizedIterable[int] = ...,
    longest: bool = ...,
    fillvalue: _U = ...,
) -> Iterator[tuple[_T | _U, ...]]: ...

class UnequalIterablesError(ValueError):
    def __init__(self, details: tuple[int, int, int] | None = ...) -> None: ...

@overload
def zip_equal(__iter1: Iterable[_T1]) -> Iterator[tuple[_T1]]: ...
@overload
def zip_equal(
    __iter1: Iterable[_T1], __iter2: Iterable[_T2]
) -> Iterator[tuple[_T1, _T2]]: ...
@overload
def zip_equal(
    __iter1: Iterable[_T],
    __iter2: Iterable[_T],
    __iter3: Iterable[_T],
    *iterables: Iterable[_T],
) -> Iterator[tuple[_T, ...]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T1],
    *,
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: None = None,
) -> Iterator[tuple[_T1 | None]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T1],
    __iter2: Iterable[_T2],
    *,
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: None = None,
) -> Iterator[tuple[_T1 | None, _T2 | None]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T],
    __iter2: Iterable[_T],
    __iter3: Iterable[_T],
    *iterables: Iterable[_T],
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: None = None,
) -> Iterator[tuple[_T | None, ...]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T1],
    *,
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: _U,
) -> Iterator[tuple[_T1 | _U]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T1],
    __iter2: Iterable[_T2],
    *,
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: _U,
) -> Iterator[tuple[_T1 | _U, _T2 | _U]]: ...
@overload
def zip_offset(
    __iter1: Iterable[_T],
    __iter2: Iterable[_T],
    __iter3: Iterable[_T],
    *iterables: Iterable[_T],
    offsets: _SizedIterable[int],
    longest: bool = ...,
    fillvalue: _U,
) -> Iterator[tuple[_T | _U, ...]]: ...
def sort_together(
    iterables: Iterable[Iterable[_T]],
    key_list: Iterable[int] = ...,
    key: Callable[..., Any] | None = ...,
    reverse: bool = ...,
) -> list[tuple[_T, ...]]: ...
def unzip(iterable: Iterable[Sequence[_T]]) -> tuple[Iterator[_T], ...]: ...
def divide(n: int, iterable: Iterable[_T]) -> list[Iterator[_T]]: ...
def always_iterable(
    obj: object,
    base_type: type | tuple[type | tuple[Any, ...], ...] | None = ...,
) -> Iterator[Any]: ...
def adjacent(
    predicate: Callable[[_T], bool],
    iterable: Iterable[_T],
    distance: int = ...,
) -> Iterator[tuple[bool, _T]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: None = None,
    valuefunc: None = None,
    reducefunc: None = None,
) -> Iterator[tuple[_T, Iterator[_T]]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: None,
    reducefunc: None,
) -> Iterator[tuple[_U, Iterator[_T]]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: None,
    valuefunc: Callable[[_T], _V],
    reducefunc: None,
) -> Iterable[tuple[_T, Iterable[_V]]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: Callable[[_T], _V],
    reducefunc: None,
) -> Iterable[tuple[_U, Iterator[_V]]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: None,
    valuefunc: None,
    reducefunc: Callable[[Iterator[_T]], _W],
) -> Iterable[tuple[_T, _W]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: None,
    reducefunc: Callable[[Iterator[_T]], _W],
) -> Iterable[tuple[_U, _W]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: None,
    valuefunc: Callable[[_T], _V],
    reducefunc: Callable[[Iterable[_V]], _W],
) -> Iterable[tuple[_T, _W]]: ...
@overload
def groupby_transform(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: Callable[[_T], _V],
    reducefunc: Callable[[Iterable[_V]], _W],
) -> Iterable[tuple[_U, _W]]: ...

class numeric_range(Generic[_T, _U], Sequence[_T], Hashable, Reversible[_T]):
    @overload
    def __init__(self, __stop: _T) -> None: ...
    @overload
    def __init__(self, __start: _T, __stop: _T) -> None: ...
    @overload
    def __init__(self, __start: _T, __stop: _T, __step: _U) -> None: ...
    def __bool__(self) -> bool: ...
    def __contains__(self, elem: object) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    @overload
    def __getitem__(self, key: int) -> _T: ...
    @overload
    def __getitem__(self, key: slice) -> numeric_range[_T, _U]: ...
    def __hash__(self) -> int: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __len__(self) -> int: ...
    def __reduce__(
        self,
    ) -> tuple[Type[numeric_range[_T, _U]], tuple[_T, _T, _U]]: ...
    def __repr__(self) -> str: ...
    def __reversed__(self) -> Iterator[_T]: ...
    def count(self, value: _T) -> int: ...
    def index(self, value: _T) -> int: ...  # type: ignore

def count_cycle(
    iterable: Iterable[_T], n: int | None = ...
) -> Iterable[tuple[int, _T]]: ...
def mark_ends(
    iterable: Iterable[_T],
) -> Iterable[tuple[bool, bool, _T]]: ...
def locate(
    iterable: Iterable[_T],
    pred: Callable[..., Any] = ...,
    window_size: int | None = ...,
) -> Iterator[int]: ...
def lstrip(
    iterable: Iterable[_T], pred: Callable[[_T], object]
) -> Iterator[_T]: ...
def rstrip(
    iterable: Iterable[_T], pred: Callable[[_T], object]
) -> Iterator[_T]: ...
def strip(
    iterable: Iterable[_T], pred: Callable[[_T], object]
) -> Iterator[_T]: ...

class islice_extended(Generic[_T], Iterator[_T]):
    def __init__(self, iterable: Iterable[_T], *args: int | None) -> None: ...
    def __iter__(self) -> islice_extended[_T]: ...
    def __next__(self) -> _T: ...
    def __getitem__(self, index: slice) -> islice_extended[_T]: ...

def always_reversible(iterable: Iterable[_T]) -> Iterator[_T]: ...
def consecutive_groups(
    iterable: Iterable[_T], ordering: Callable[[_T], int] = ...
) -> Iterator[Iterator[_T]]: ...
@overload
def difference(
    iterable: Iterable[_T],
    func: Callable[[_T, _T], _U] = ...,
    *,
    initial: None = ...,
) -> Iterator[_T | _U]: ...
@overload
def difference(
    iterable: Iterable[_T], func: Callable[[_T, _T], _U] = ..., *, initial: _U
) -> Iterator[_U]: ...

class SequenceView(Generic[_T], Sequence[_T]):
    def __init__(self, target: Sequence[_T]) -> None: ...
    @overload
    def __getitem__(self, index: int) -> _T: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[_T]: ...
    def __len__(self) -> int: ...

class seekable(Generic[_T], Iterator[_T]):
    def __init__(
        self, iterable: Iterable[_T], maxlen: int | None = ...
    ) -> None: ...
    def __iter__(self) -> seekable[_T]: ...
    def __next__(self) -> _T: ...
    def __bool__(self) -> bool: ...
    @overload
    def peek(self) -> _T: ...
    @overload
    def peek(self, default: _U) -> _T | _U: ...
    def elements(self) -> SequenceView[_T]: ...
    def seek(self, index: int) -> None: ...
    def relative_seek(self, count: int) -> None: ...

class run_length:
    @staticmethod
    def encode(iterable: Iterable[_T]) -> Iterator[tuple[_T, int]]: ...
    @staticmethod
    def decode(iterable: Iterable[tuple[_T, int]]) -> Iterator[_T]: ...

def exactly_n(
    iterable: Iterable[_T], n: int, predicate: Callable[[_T], object] = ...
) -> bool: ...
def circular_shifts(iterable: Iterable[_T]) -> list[tuple[_T, ...]]: ...
def make_decorator(
    wrapping_func: Callable[..., _U], result_index: int = ...
) -> Callable[..., Callable[[Callable[..., Any]], Callable[..., _U]]]: ...
@overload
def map_reduce(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: None = ...,
    reducefunc: None = ...,
) -> dict[_U, list[_T]]: ...
@overload
def map_reduce(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: Callable[[_T], _V],
    reducefunc: None = ...,
) -> dict[_U, list[_V]]: ...
@overload
def map_reduce(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: None = ...,
    reducefunc: Callable[[list[_T]], _W] = ...,
) -> dict[_U, _W]: ...
@overload
def map_reduce(
    iterable: Iterable[_T],
    keyfunc: Callable[[_T], _U],
    valuefunc: Callable[[_T], _V],
    reducefunc: Callable[[list[_V]], _W],
) -> dict[_U, _W]: ...
def rlocate(
    iterable: Iterable[_T],
    pred: Callable[..., object] = ...,
    window_size: int | None = ...,
) -> Iterator[int]: ...
def replace(
    iterable: Iterable[_T],
    pred: Callable[..., object],
    substitutes: Iterable[_U],
    count: int | None = ...,
    window_size: int = ...,
) -> Iterator[_T | _U]: ...
def partitions(iterable: Iterable[_T]) -> Iterator[list[list[_T]]]: ...
def set_partitions(
    iterable: Iterable[_T], k: int | None = ...
) -> Iterator[list[list[_T]]]: ...

class time_limited(Generic[_T], Iterator[_T]):
    def __init__(
        self, limit_seconds: float, iterable: Iterable[_T]
    ) -> None: ...
    def __iter__(self) -> islice_extended[_T]: ...
    def __next__(self) -> _T: ...

@overload
def only(
    iterable: Iterable[_T], *, too_long: _Raisable | None = ...
) -> _T | None: ...
@overload
def only(
    iterable: Iterable[_T], default: _U, too_long: _Raisable | None = ...
) -> _T | _U: ...
def ichunked(iterable: Iterable[_T], n: int) -> Iterator[Iterator[_T]]: ...
def distinct_combinations(
    iterable: Iterable[_T], r: int
) -> Iterator[tuple[_T, ...]]: ...
def filter_except(
    validator: Callable[[Any], object],
    iterable: Iterable[_T],
    *exceptions: Type[BaseException],
) -> Iterator[_T]: ...
def map_except(
    function: Callable[[Any], _U],
    iterable: Iterable[_T],
    *exceptions: Type[BaseException],
) -> Iterator[_U]: ...
def map_if(
    iterable: Iterable[Any],
    pred: Callable[[Any], bool],
    func: Callable[[Any], Any],
    func_else: Callable[[Any], Any] | None = ...,
) -> Iterator[Any]: ...
def sample(
    iterable: Iterable[_T],
    k: int,
    weights: Iterable[float] | None = ...,
) -> list[_T]: ...
def is_sorted(
    iterable: Iterable[_T],
    key: Callable[[_T], _U] | None = ...,
    reverse: bool = False,
    strict: bool = False,
) -> bool: ...

class AbortThread(BaseException):
    pass

class callback_iter(Generic[_T], Iterator[_T]):
    def __init__(
        self,
        func: Callable[..., Any],
        callback_kwd: str = ...,
        wait_seconds: float = ...,
    ) -> None: ...
    def __enter__(self) -> callback_iter[_T]: ...
    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None: ...
    def __iter__(self) -> callback_iter[_T]: ...
    def __next__(self) -> _T: ...
    def _reader(self) -> Iterator[_T]: ...
    @property
    def done(self) -> bool: ...
    @property
    def result(self) -> Any: ...

def windowed_complete(
    iterable: Iterable[_T], n: int
) -> Iterator[tuple[_T, ...]]: ...
def all_unique(
    iterable: Iterable[_T], key: Callable[[_T], _U] | None = ...
) -> bool: ...
def nth_product(index: int, *args: Iterable[_T]) -> tuple[_T, ...]: ...
def nth_combination_with_replacement(
    iterable: Iterable[_T], r: int, index: int
) -> tuple[_T, ...]: ...
def nth_permutation(
    iterable: Iterable[_T], r: int, index: int
) -> tuple[_T, ...]: ...
def value_chain(*args: _T | Iterable[_T]) -> Iterable[_T]: ...
def product_index(element: Iterable[_T], *args: Iterable[_T]) -> int: ...
def combination_index(
    element: Iterable[_T], iterable: Iterable[_T]
) -> int: ...
def combination_with_replacement_index(
    element: Iterable[_T], iterable: Iterable[_T]
) -> int: ...
def permutation_index(
    element: Iterable[_T], iterable: Iterable[_T]
) -> int: ...
def repeat_each(iterable: Iterable[_T], n: int = ...) -> Iterator[_T]: ...

class countable(Generic[_T], Iterator[_T]):
    def __init__(self, iterable: Iterable[_T]) -> None: ...
    def __iter__(self) -> countable[_T]: ...
    def __next__(self) -> _T: ...
    items_seen: int

def chunked_even(iterable: Iterable[_T], n: int) -> Iterator[list[_T]]: ...
def zip_broadcast(
    *objects: _T | Iterable[_T],
    scalar_types: type | tuple[type | tuple[Any, ...], ...] | None = ...,
    strict: bool = ...,
) -> Iterable[tuple[_T, ...]]: ...
def unique_in_window(
    iterable: Iterable[_T], n: int, key: Callable[[_T], _U] | None = ...
) -> Iterator[_T]: ...
def duplicates_everseen(
    iterable: Iterable[_T], key: Callable[[_T], _U] | None = ...
) -> Iterator[_T]: ...
def duplicates_justseen(
    iterable: Iterable[_T], key: Callable[[_T], _U] | None = ...
) -> Iterator[_T]: ...
def classify_unique(
    iterable: Iterable[_T], key: Callable[[_T], _U] | None = ...
) -> Iterator[tuple[_T, bool, bool]]: ...

class _SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool: ...

_SupportsLessThanT = TypeVar("_SupportsLessThanT", bound=_SupportsLessThan)

@overload
def minmax(
    iterable_or_value: Iterable[_SupportsLessThanT], *, key: None = None
) -> tuple[_SupportsLessThanT, _SupportsLessThanT]: ...
@overload
def minmax(
    iterable_or_value: Iterable[_T], *, key: Callable[[_T], _SupportsLessThan]
) -> tuple[_T, _T]: ...
@overload
def minmax(
    iterable_or_value: Iterable[_SupportsLessThanT],
    *,
    key: None = None,
    default: _U,
) -> _U | tuple[_SupportsLessThanT, _SupportsLessThanT]: ...
@overload
def minmax(
    iterable_or_value: Iterable[_T],
    *,
    key: Callable[[_T], _SupportsLessThan],
    default: _U,
) -> _U | tuple[_T, _T]: ...
@overload
def minmax(
    iterable_or_value: _SupportsLessThanT,
    __other: _SupportsLessThanT,
    *others: _SupportsLessThanT,
) -> tuple[_SupportsLessThanT, _SupportsLessThanT]: ...
@overload
def minmax(
    iterable_or_value: _T,
    __other: _T,
    *others: _T,
    key: Callable[[_T], _SupportsLessThan],
) -> tuple[_T, _T]: ...
def longest_common_prefix(
    iterables: Iterable[Iterable[_T]],
) -> Iterator[_T]: ...
def iequals(*iterables: Iterable[Any]) -> bool: ...
def constrained_batches(
    iterable: Iterable[_T],
    max_size: int,
    max_count: int | None = ...,
    get_len: Callable[[_T], object] = ...,
    strict: bool = ...,
) -> Iterator[tuple[_T]]: ...
def gray_product(*iterables: Iterable[_T]) -> Iterator[tuple[_T, ...]]: ...
def partial_product(*iterables: Iterable[_T]) -> Iterator[tuple[_T, ...]]: ...
def takewhile_inclusive(
    predicate: Callable[[_T], bool], iterable: Iterable[_T]
) -> Iterator[_T]: ...
def outer_product(
    func: Callable[[_T, _U], _V],
    xs: Iterable[_T],
    ys: Iterable[_U],
    *args: Any,
    **kwargs: Any,
) -> Iterator[tuple[_V, ...]]: ...
def iter_suppress(
    iterable: Iterable[_T],
    *exceptions: Type[BaseException],
) -> Iterator[_T]: ...
def filter_map(
    func: Callable[[_T], _V | None],
    iterable: Iterable[_T],
) -> Iterator[_V]: ...
def powerset_of_sets(iterable: Iterable[_T]) -> Iterator[set[_T]]: ...
def join_mappings(
    **field_to_map: Mapping[_T, _V]
) -> dict[_T, dict[str, _V]]: ...
def doublestarmap(
    func: Callable[..., _T],
    iterable: Iterable[Mapping[str, Any]],
) -> Iterator[_T]: ...
def dft(xarr: Sequence[complex]) -> Iterator[complex]: ...
def idft(Xarr: Sequence[complex]) -> Iterator[complex]: ...



import sysactions
from msgbus import TK_SYS_ACK
from bustoken import TK_REMOTE_ORDER_GET_POS_ID
from manager.reports import cashReport
from .. import mb_context


@sysactions.action
def do_delivery_report(pos_id):
    has_delivery = mb_context.MB_LocateService(mb_context.hv_service, 'RemoteOrder', maxretries=1)
    if has_delivery:
        msg = mb_context.MB_EasySendMessage('RemoteOrder', TK_REMOTE_ORDER_GET_POS_ID, data='', timeout=5000*1000)
        if msg.token == TK_SYS_ACK:
            remote_pos_id = msg.data
            cashReport(
                pos_id,
                ask_operator="notAsk",
                date_type="RealDate",
                ask_pos=int(remote_pos_id),
                is_integrated_delivery=True
            )
    return True




import os
import re


_default_seps = os.sep + str(os.altsep) * bool(os.altsep)


class Translator:
    """
    >>> Translator('xyz')
    Traceback (most recent call last):
    ...
    AssertionError: Invalid separators

    >>> Translator('')
    Traceback (most recent call last):
    ...
    AssertionError: Invalid separators
    """

    seps: str

    def __init__(self, seps: str = _default_seps):
        assert seps and set(seps) <= set(_default_seps), "Invalid separators"
        self.seps = seps

    def translate(self, pattern):
        """
        Given a glob pattern, produce a regex that matches it.
        """
        return self.extend(self.translate_core(pattern))

    def extend(self, pattern):
        r"""
        Extend regex for pattern-wide concerns.

        Apply '(?s:)' to create a non-matching group that
        matches newlines (valid on Unix).

        Append '\Z' to imply fullmatch even when match is used.
        """
        return rf'(?s:{pattern})\Z'

    def translate_core(self, pattern):
        r"""
        Given a glob pattern, produce a regex that matches it.

        >>> t = Translator()
        >>> t.translate_core('*.txt').replace('\\\\', '')
        '[^/]*\\.txt'
        >>> t.translate_core('a?txt')
        'a[^/]txt'
        >>> t.translate_core('**/*').replace('\\\\', '')
        '.*/[^/][^/]*'
        """
        self.restrict_rglob(pattern)
        return ''.join(map(self.replace, separate(self.star_not_empty(pattern))))

    def replace(self, match):
        """
        Perform the replacements for a match from :func:`separate`.
        """
        return match.group('set') or (
            re.escape(match.group(0))
            .replace('\\*\\*', r'.*')
            .replace('\\*', rf'[^{re.escape(self.seps)}]*')
            .replace('\\?', r'[^/]')
        )

    def restrict_rglob(self, pattern):
        """
        Raise ValueError if ** appears in anything but a full path segment.

        >>> Translator().translate('**foo')
        Traceback (most recent call last):
        ...
        ValueError: ** must appear alone in a path segment
        """
        seps_pattern = rf'[{re.escape(self.seps)}]+'
        segments = re.split(seps_pattern, pattern)
        if any('**' in segment and segment != '**' for segment in segments):
            raise ValueError("** must appear alone in a path segment")

    def star_not_empty(self, pattern):
        """
        Ensure that * will not match an empty segment.
        """

        def handle_segment(match):
            segment = match.group(0)
            return '?*' if segment == '*' else segment

        not_seps_pattern = rf'[^{re.escape(self.seps)}]+'
        return re.sub(not_seps_pattern, handle_segment, pattern)


def separate(pattern):
    """
    Separate out character sets to avoid translating their contents.

    >>> [m.group(0) for m in separate('*.txt')]
    ['*.txt']
    >>> [m.group(0) for m in separate('a[?]txt')]
    ['a', '[?]', 'txt']
    """
    return re.finditer(r'([^\[]+)|(?P<set>[\[].*?[\]])|([\[][^\]]*$)', pattern)



# -*- coding: utf-8 -*-
import datetime
import time

import msgbus
import persistence
import sysactions
from xml.etree import cElementTree as eTree

from actions.models import AuthorizationResponse
from actions.models.mw_sys_errors import MwSysErrors
from actions.util import (
    get_authorization,
    get_operator_info,
    get_user_level,
)
from api.order import get_order_service
from model.customexception import CloseOrderByOperatorCloseException
from typing import (
    Dict,
    Optional,
)
from .. import (
    logger,
    mb_context,
    pos_config,
)


def get_pos_service(pos_id):
    return "POS%d" % int(pos_id)


def login_user(pos_id, user_id):
    try:
        long_name, level, user_id = get_user_info(user_id)

        msg = sysactions.send_message(
            get_pos_service(pos_id),
            msgbus.TK_POS_USERLOGIN,
            msgbus.FM_PARAM, "{}\0{}\0{}".format(pos_id, user_id, long_name),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            model = sysactions.get_model(pos_id)
            message = sysactions.translate_message(model, "ERROR") + ": "
            error = MwSysErrors.get_name(msg.data.replace("\0", ""))
            _handle_current_order_error(error, model, pos_id)
            message += sysactions.translate_message(model, error)
            sysactions.show_messagebox(pos_id, message, "$INFORMATION")
            logger.error("Invalid response from PosCtrl while login user: {} - {}".format(msg.token, msg.data))
            return False

    except (Exception, BaseException):
        logger.error("Exception while login user")
        sysactions.show_messagebox(pos_id, "$CANNOT_LOGIN_OPERATOR")
        return False

    return True


def _handle_current_order_error(error, model, pos_id):
    # type: (str, eTree.XML, int) -> None

    try:
        if error == "SE_ORDEROPENED":
            order_service = get_order_service(pos_id=pos_id)
            order_service.void_order()
    except (Exception,):
        logger.exception("Error voiding order on user login")


def open_user(pos_id, user_id):
    msg = None
    try:
        long_name, level, user_id = get_user_info(user_id)

        msg = sysactions.send_message(
            get_pos_service(pos_id),
            msgbus.TK_POS_USEROPEN,
            msgbus.FM_PARAM,
            "%s\0%s\0%s\0%s\0%s" % (pos_id, user_id, 0.0, long_name, level),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            if _user_is_already_opened(msg):
                return True

            logger.exception("Error opening user: Token: {} / Data: {}".format(msg.token, msg.data))
            sysactions.show_messagebox(pos_id, "$CANNOT_OPEN_OPERATOR")
            return False
    except (Exception, BaseException):
        logger.exception("Error opening user {} on POS {}".format(user_id, pos_id))
        message = "$CANNOT_OPEN_OPERATOR"
        has_error = msg is not None and msg.data and len(msg.data.split("\0")) > 1
        if has_error and msg.data.split("\0")[0] in MwSysErrors.list_values():
            model = sysactions.get_model(pos_id)
            message = sysactions.translate_message(model, MwSysErrors.get_name(msg.data.split("\0")[0]))

        sysactions.show_messagebox(pos_id, message)
        return False
    return True


def _user_is_already_opened(msg):
    return msg.data.startswith("5007")


def close_user(
    pos_id,                     # type: str
    user_id,                    # type: str
    pos_to_close_user=None,     # type: Optional[str]
):
    # type: (...) -> bool

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    try:
        msg = sysactions.send_message(
            get_pos_service(pos_id=pos_to_close_user),
            msgbus.TK_POS_USERLOGOUT,
            msgbus.FM_PARAM,
            "%s\0%s" % (pos_to_close_user, user_id),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            if _user_is_already_closed(msg):
                return True

            logger.exception("Error closing user: Token: {} / Data: {}".format(msg.token, msg.data))
            sysactions.show_messagebox(pos_id, "$CANNOT_CLOSE_OPERATOR")
            return False
    except (Exception, BaseException):
        logger.exception("Error closing user {} on POS {}".format(user_id, pos_to_close_user))
        sysactions.show_messagebox(pos_id, "$CANNOT_CLOSE_OPERATOR")
        return False

    return True


def _user_is_already_closed(msg):
    return msg.data.startswith("5007")


def check_current_day(pos_id):
    model = sysactions.get_model(pos_id)
    if sysactions.is_day_blocked(model):
        sysactions.show_messagebox(pos_id, "$POS_IS_BLOCKED_BY_TIME", "$INFORMATION")
        return False

    if not sysactions.is_day_opened(model):
        sysactions.show_messagebox(pos_id, "$NEED_TO_OPEN_BDAY_FIRST", "$INFORMATION")
        return False

    return True


def get_user_info(user_id):
    msg = sysactions.send_message("UserControl", msgbus.TK_USERCTRL_GETINFO, msgbus.FM_PARAM, user_id)
    if msg.token != msgbus.TK_SYS_ACK:
        raise Exception("Error getting user information {}-{}".format(msg.token, msg.data))

    user_info = eTree.XML(msg.data)
    long_name = user_info.find("user").get("LongName")
    level = user_info.find("user").get("Level")
    user_id = user_info.find("user").get("UserId")

    return long_name, level, user_id


def print_open_day_report(pos_id, initial_float):
    model = sysactions.get_model(pos_id)
    operator = sysactions.get_current_operator(model)
    operator_name = (operator.get("name")).encode('utf-8')
    operator_id = operator.get("id")
    period = sysactions.get_business_period(model)
    authorized_by = sysactions.get_custom(model, 'Last Manager ID', operator_id)

    sysactions.print_report(
        pos_id,
        model,
        True,
        "login_operator_report",
        pos_id,
        operator_id,
        operator_name,
        initial_float,
        period,
        authorized_by
    )


def get_pod_type_and_pos_function(pos_id):
    pod_type = sysactions.get_cfg(pos_id).key_value("pod") or 'FL'
    pos_function_config = sysactions.get_cfg(pos_id).key_value("posFunction") or 'FL'
    pos_function_list = pos_function_config.split("|")
    pos_function = pos_function_list[0]
    if len(pos_function_list) > 1:
        pos_function_index = sysactions.show_listbox(pos_id, pos_function_list, message="$SELECT_A_POS_FUNCTION")

        if pos_function_index is None:
            return None, None

        pos_function = pos_function_list[pos_function_index]

    return pod_type, pos_function


def change_pod_type_and_pos_function(pos_id, pod_type, pos_function):
    if pos_function == "DS":
        sysactions.check_main_screen(pos_id)
    sysactions.send_message(
        get_pos_service(pos_id),
        msgbus.TK_POS_SETPOD,
        msgbus.FM_PARAM, "{}\0{}\0".format(pos_id, pod_type),
        timeout=60 * 1000000)
    sysactions.send_message(
        get_pos_service(pos_id),
        msgbus.TK_POS_SETFUNCTION,
        msgbus.FM_PARAM, "{}\0{}\0".format(pos_id, pos_function),
        timeout=60 * 1000000)


def get_initial_float(
    pos_id,         # type: str
    pod_type,       # type: str
    pos_function,   # type: str
):
    # type: (...) -> float

    if pod_type != "OT" and pos_function != "OT":
        index = None
        list_min_values_drawer = sysactions.get_storewide_config("Store.MinValuesDrawer")
        if list_min_values_drawer:
            initial_float = _get_initial_float_by_min_values(
                pos_id=pos_id,
                list_min_values_drawer=list_min_values_drawer,
                index=index,
            )
        else:
            initial_float = _get_initial_float_by_user_input(pos_id=pos_id)

        return round(float(initial_float or 0.0), 2)
    else:
        return 0.0


def _get_initial_float_by_user_input(
    pos_id,     # type: str
):
    # type: (...) -> Optional[str]

    initial_float = sysactions.show_keyboard(
        posid=pos_id,
        message="$ENTER_THE_INITIAL_FLOAT_AMOUNT",
        title="$OPERATOR_OPENING",
        mask="CURRENCY",
        numpad=True,
        timeout=720000,
    )
    if initial_float is None:
        raise sysactions.StopAction()

    return initial_float


def _get_initial_float_by_min_values(
    pos_id,                     # type: str
    list_min_values_drawer,     # type: str
    index,                      # type: Optional[int]
):
    # type: (...) -> str

    list_limits = [("R$%.2f" % float(x)).replace(".", ",") for x in list_min_values_drawer.split(";")]
    if len(list_limits) > 1:
        index = sysactions.show_listbox(
            posid=pos_id,
            options=list_limits,
            message="$ENTER_THE_INITIAL_FLOAT_AMOUNT",
            title="$OPERATOR_LOGIN",
            icon="info",
            timeout=720000,
        )

    elif len(list_limits) == 1:
        index = 0

    if index is None:
        raise sysactions.StopAction()

    initial_float = list_min_values_drawer.split(";")[index]
    return initial_float


def save_initial_float(
    initial_float,  # type: float
    pos_id,         # type: str
    user_id,        # type: str
):
    # type: (...) -> None

    try:
        wait_user_logged_in(pos_id, user_id)
        model = sysactions.get_model(pos_id)
        _, session = get_operator_info(model, user_id)
        sql = "insert into Transfer (Period, PosId, SessionId, Type, Description, Amount, TenderId) " \
              "values ({}, {}, '{}', 1, 'INITIAL_AMOUNT', {}, 0);"\
            .format(datetime.datetime.now().strftime('%Y%m%d'), pos_id, session, initial_float)
        persistence.Driver().open(mb_context, pos_id).query(sql)

        _send_initial_amount_audit_event(
            pos_id=pos_id,
            model=model,
            session=session,
            initial_float=initial_float,
        )
    except (Exception,):
        logger.exception("Fail set initial float to user: {}".format(user_id))


def _send_initial_amount_audit_event(
    pos_id,         # type: str
    model,          # type: eTree.Element
    session,        # type: str
    initial_float,  # type: float
):
    # type: (...) -> None

    timestamp = time.gmtime(time.time())
    timestamp = time.strftime("%Y-%m-%dT%H:%M:%SZ", timestamp)
    period = sysactions.get_business_period(model=model)

    initial_amount_xml = eTree.Element(
        "InitialAmount",
        storeId=pos_config.store_id,
        period=period,
        posId=pos_id,
        sessionId=session,
        amount=str(initial_float),
        tenderId="0",
        timestamp=timestamp,
    )

    mb_context.MB_EasyEvtSend(
        subject="INITIAL_AMOUNT",
        type="INITIAL_AMOUNT",
        xml=eTree.tostring(initial_amount_xml),
    )


def wait_user_logged_in(pos_id, user_id):
    # type: (str, str) -> None

    time_limit = (time.time() + 1.5)
    while (time.time() < time_limit) and (not has_session(sysactions.get_model(pos_id), user_id)):
        time.sleep(0.1)


def has_session(model, user_id):
    # type: (eTree, str) -> bool

    _, session = get_operator_info(model, user_id)
    return session not in ['', None]


def user_inserted_initial_amount(pos_id, session):
    try:
        if session in ['', None]:
            return False

        sql = "select count(*) from Transfer " \
              "where SessionId = '{}' and Description = 'INITIAL_AMOUNT'".\
            format(session)
        conn = persistence.Driver().open(mb_context, pos_id)
        row = conn.select(sql).get_row(0)
        return int(row.get_entry(0)) > 0
    except (Exception,):
        return False


def get_authorization_if_necessary(
    model,      # type: eTree.XML
    pos_id,     # type: str
    user_info,  # type: Dict
    min_level   # type: int
):
    # type: (...) -> AuthorizationResponse

    user_id = user_info["UserId"]
    user_name = user_info["UserName"]
    user_level = int(user_info["Level"] or 0)
    if user_level < min_level:
        response = get_authorization(pos_id, min_level=min_level, model=model)
        if not response[0]:
            return AuthorizationResponse(False, "", "")

        user_name = response[1]

    return AuthorizationResponse(
        success=True,
        authorized_by=user_name,
        authorized_by_level=get_user_level(user_id=user_id),
    )


def save_last_authorization(
    pos_id,                 # type: str
    authorized_by,          # type: str
    authorized_by_level,    # type: str
):
    # type: (...) -> None

    sysactions.set_custom(pos_id, 'Last Manager ID', authorized_by, True)
    sysactions.set_custom(pos_id, 'Authorization Level', authorized_by_level, True)


def validate_user_level(
    pos_id,     # type: str
    user_info,  # type: Dict
):
    # type: (...) -> None

    if int(user_info.get("Level", -1)) < 0:
        logger.error("Access Denied for user: {}".format(user_info["UserName"]))
        sysactions.show_messagebox(pos_id, "$ACCESS_DENIED", "$INFORMATION")
        raise sysactions.StopAction()


def request_operator_confirmation(
    pos_id,             # type: str
    pos_to_close_user,  # type: str
):
    # type: (...) -> None

    wait_dialog_id = sysactions.show_messagebox(
        posid=pos_id,
        message="$WAITING_FOR_OPERATOR_CONFIRMATION",
        buttons="",
        asynch=True,
    )
    confirmation = sysactions.show_confirmation(
        posid=pos_to_close_user,
        message="$CONFIRM_CLOSE_USER_FROM_ANOTHER_POS",
        icon="warning",
        timeout=10000,
    )
    sysactions.close_asynch_dialog(
        posid=pos_id,
        dlgid=wait_dialog_id,
    )
    if confirmation is False:
        sysactions.show_messagebox(
            posid=pos_id,
            message="$OPERATION_CANCELLED_BY_OPERATOR",
            timeout=5000,
        )
        raise sysactions.StopAction()


def close_current_order(
    model,                  # type: eTree.Element
    current_pos_id,         # type: str
    pos_to_close_order,     # type: str
    user_id=None,           # type: Optional[str]
):
    # type: (...) -> bool

    logger.info("[Close order from another POS] Init process can_close_order_from_another_pos")
    current_order = sysactions.get_current_order(model=model)
    if current_order is not None:
        from actions.login.common_close_user import (
            close_quick_service_order_from_another_pos,
            close_table_service_from_another_pos,
        )

        is_quick_service_order = _order_is_quick_service(order=current_order)
        if pos_to_close_order is not None:
            sysactions.show_messagebox(
                posid=pos_to_close_order,
                message="$PLEASE_WAIT",
                timeout="$INFO",
                asynch=True,
                buttons=""
            )

        try:
            if is_quick_service_order:
                close_quick_service_order_from_another_pos(
                    pos_id=current_pos_id,
                    order=current_order,
                    pos_to_close_user=pos_to_close_order,
                    user_id=user_id,
                    must_store_quick_service_order=True,
                )
            else:
                close_table_service_from_another_pos(
                    pos_id=current_pos_id,
                    order=current_order,
                    pos_to_close_user=pos_to_close_order,
                    user_id=user_id,
                    ask_to_void_order=False,
                )
        except (CloseOrderByOperatorCloseException, Exception):
            logger.exception("[Close order from another POS] Cannot complete Close Order")
            return False
        finally:
            sysactions.close_asynch_dialog(pos_to_close_order, -1)

    logger.info("[Close order from another POS] End process can_close_order_from_another_pos")

    return True


def _order_is_quick_service(
    order,  # type: eTree.Element
):
    # type: (...) -> bool

    table_id = order.find("CustomOrderProperties/OrderProperty/[@key='TABLE_ID']")
    return table_id is None



from typing import List

from boss.interactor.get_info.dto import GetInfoRequest
from boss.model import User
from boss.util import convert_user_dto_into_user_response
from messagebus import (
    AckMessage,
    Event,
    Message,
    MessageBus,
)
from ._BaseProcessor import BaseProcessor


class GetUsersListProcessor(BaseProcessor):

    def parse_data(
        self,
        data: bytes,
    ) -> None:

        return None

    def format_response(
        self,
        message_bus: MessageBus,
        message: Message,
        event: Event,
        request: GetInfoRequest,
        result: List[User],
    ) -> None:

        if result is None:
            reply_message = AckMessage(data=b"<UserInfo/>")
        else:
            sorted_result = sorted(result, key=lambda x: x.user_id)
            reply_data = "<UserInfo>"
            for user_dto in sorted_result:
                user_data = convert_user_dto_into_user_response(user=user_dto)
                reply_data += "{}".format(user_data.get_info_response())
            reply_data += "</UserInfo>"

            reply_message = AckMessage(data=reply_data.encode())

        message_bus.reply_message(
            message=message,
            reply_message=reply_message,
        )



from logging import Logger
from typing import (
    Optional,
    List,
)

from production.model import (
    ProductionOrder,
    ProdStates,
)
from .._OrderChangerProductionBox import OrderChangerProductionBox


class StoredOrderUnsplitter(OrderChangerProductionBox):

    def __init__(
        self,                       # type: StoredOrderUnsplitter
        name,                       # type: str
        sons,                       # type: object
        prod_states_clean_cache,    # type: List[str]
        logger=None,                # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(StoredOrderUnsplitter, self).__init__(name, sons, logger)

        self.prod_states_clean_cache = prod_states_clean_cache

    def change_order(
        self,     # type: StoredOrderUnsplitter
        order,    # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        if not order.stored_split_order:
            return order

        cloned_order = self._get_cloned_unsplitted_order(order=order)
        cached_order = self.get_order(order_id=cloned_order.order_id)
        if cached_order is None:
            self.save_order(order=cloned_order)
        else:
            self._update_cached_order(
                cloned_order=cloned_order,
                cached_order=cached_order,
            )

        updated_order = self.get_order(order_id=cloned_order.order_id)
        if updated_order.prod_state in self.prod_states_clean_cache:
            self.delete_order(order_id=updated_order.order_id)

        return updated_order

    def _update_cached_order(
        self,            # type: StoredOrderUnsplitter
        cloned_order,    # type: ProductionOrder
        cached_order,    # type: ProductionOrder
    ):
        # type: (...) -> None

        if cloned_order.prod_state == ProdStates.INVALID and len(cloned_order.items) == 0:
            return

        self._update_cached_items(
            cloned_order=cloned_order,
            cached_order=cached_order,
        )

        if cached_order.state != cloned_order.state or cached_order.prod_state != cloned_order.prod_state:
            cached_order.items.extend(cloned_order.items)
            cloned_order.items = cached_order.items
            self.save_order(order=cloned_order)
        else:
            cached_order.items.extend(cloned_order.items)
            self.save_order(order=cached_order)

    @staticmethod
    def _update_cached_items(
        cloned_order,    # type: ProductionOrder
        cached_order,    # type: ProductionOrder
    ):
        # type: (...) -> None

        items_to_remove_from_cache = []
        for cloned_item in cloned_order.items:
            for cached_item in cached_order.items:
                if cloned_item.get_line_id() == cached_item.get_line_id():
                    items_to_remove_from_cache.append(cached_item)
                    break

        for item_to_remove in items_to_remove_from_cache:
            cached_order.items.remove(item_to_remove)

    @staticmethod
    def _get_cloned_unsplitted_order(
        order,    # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        order_cloned = order.clone(items=order.items)
        split_indexes = order.order_id.split("-")
        if len(split_indexes) == 1:
            order_cloned.order_id = split_indexes[0]
        else:
            order_cloned.order_id = "-".join(split_indexes[:len(split_indexes) - 1])

        order_cloned.stored_split_order = False
        return order_cloned



from production.box.condition import (
    ConditionBox,
    CustomPropertyCondition,
    PodTypeCondition,
    CustomPropertyValuesCondition,
    OrderStateCondition,
    DisplayTimeCondition,
    SaleTypeCondition
)
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class ConditionBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(ConditionBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> ConditionBox

        return ConditionBox(
            config.name,
            self.get_condition(config.extra_config),
            config.sons[0],
            config.sons[1],
            self.get_logger(config)
        )

    @staticmethod
    def get_condition(config):
        condition_config = config["Condition"]
        if condition_config["Name"] == "CustomPropertyCondition":
            return CustomPropertyCondition(condition_config["PropertyName"])

        if condition_config["Name"] == "CustomPropertyValuesCondition":
            allowed_values = condition_config.get("AllowedValues")
            excluded_values = condition_config.get("ExcludedValues")
            return CustomPropertyValuesCondition(condition_config["PropertyName"], allowed_values, excluded_values)

        if condition_config["Name"] == "DisplayTimeCondition":
            return DisplayTimeCondition(condition_config["MaxDisplayTime"])

        if condition_config["Name"] == "PodTypeCondition":
            return PodTypeCondition(condition_config["ValidPodTypes"])

        if condition_config["Name"] == "OrderStateCondition":
            return OrderStateCondition(condition_config["OrderStates"])

        if condition_config["Name"] == "SaleTypeCondition":
            return SaleTypeCondition(condition_config["ValidSaleTypes"])



from logging import Logger

from mw_helper import ensure_iterable
from production.repository import ProductRepository
from typing import List, Optional, Union

from ._DoneWhenAnyDoneBox import DoneWhenAnyDoneBox
from ._ProductionBox import ProductionBox


class JitDoneWhenAnyDoneBox(DoneWhenAnyDoneBox):
    def __init__(self, name, sons, product_repository, jit_lines, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], ProductRepository, List[str], Logger) -> None

        jit_lines = ensure_iterable(jit_lines)

        part_codes = product_repository.get_part_codes_of_jit_lines(jit_lines)
        super(JitDoneWhenAnyDoneBox, self).__init__(name, sons, part_codes, logger)



from __future__ import annotations

import itertools
import os
import sys
from importlib.machinery import EXTENSION_SUFFIXES
from importlib.util import cache_from_source as _compiled_file_name
from pathlib import Path
from typing import TYPE_CHECKING, Iterator

from setuptools.dist import Distribution
from setuptools.errors import BaseError
from setuptools.extension import Extension, Library

from distutils import log
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler, get_config_var

if TYPE_CHECKING:
    # Cython not installed on CI tests, causing _build_ext to be `Any`
    from distutils.command.build_ext import build_ext as _build_ext
else:
    try:
        # Attempt to use Cython for building extensions, if available
        from Cython.Distutils.build_ext import build_ext as _build_ext

        # Additionally, assert that the compiler module will load
        # also. Ref #1229.
        __import__('Cython.Compiler.Main')
    except ImportError:
        from distutils.command.build_ext import build_ext as _build_ext

# make sure _config_vars is initialized
get_config_var("LDSHARED")
# Not publicly exposed in typeshed distutils stubs, but this is done on purpose
# See https://github.com/pypa/setuptools/pull/4228#issuecomment-1959856400
from distutils.sysconfig import _config_vars as _CONFIG_VARS  # noqa: E402


def _customize_compiler_for_shlib(compiler):
    if sys.platform == "darwin":
        # building .dylib requires additional compiler flags on OSX; here we
        # temporarily substitute the pyconfig.h variables so that distutils'
        # 'customize_compiler' uses them before we build the shared libraries.
        tmp = _CONFIG_VARS.copy()
        try:
            # XXX Help!  I don't have any idea whether these are right...
            _CONFIG_VARS['LDSHARED'] = (
                "gcc -Wl,-x -dynamiclib -undefined dynamic_lookup"
            )
            _CONFIG_VARS['CCSHARED'] = " -dynamiclib"
            _CONFIG_VARS['SO'] = ".dylib"
            customize_compiler(compiler)
        finally:
            _CONFIG_VARS.clear()
            _CONFIG_VARS.update(tmp)
    else:
        customize_compiler(compiler)


have_rtld = False
use_stubs = False
libtype = 'shared'

if sys.platform == "darwin":
    use_stubs = True
elif os.name != 'nt':
    try:
        import dl  # type: ignore[import-not-found] # https://github.com/python/mypy/issues/13002

        use_stubs = have_rtld = hasattr(dl, 'RTLD_NOW')
    except ImportError:
        pass


def if_dl(s):
    return s if have_rtld else ''


def get_abi3_suffix():
    """Return the file extension for an abi3-compliant Extension()"""
    for suffix in EXTENSION_SUFFIXES:
        if '.abi3' in suffix:  # Unix
            return suffix
        elif suffix == '.pyd':  # Windows
            return suffix
    return None


class build_ext(_build_ext):
    distribution: Distribution  # override distutils.dist.Distribution with setuptools.dist.Distribution
    editable_mode = False
    inplace = False

    def run(self):
        """Build extensions in build directory, then copy if --inplace"""
        old_inplace, self.inplace = self.inplace, 0
        _build_ext.run(self)
        self.inplace = old_inplace
        if old_inplace:
            self.copy_extensions_to_source()

    def _get_inplace_equivalent(self, build_py, ext: Extension) -> tuple[str, str]:
        fullname = self.get_ext_fullname(ext.name)
        filename = self.get_ext_filename(fullname)
        modpath = fullname.split('.')
        package = '.'.join(modpath[:-1])
        package_dir = build_py.get_package_dir(package)
        inplace_file = os.path.join(package_dir, os.path.basename(filename))
        regular_file = os.path.join(self.build_lib, filename)
        return (inplace_file, regular_file)

    def copy_extensions_to_source(self):
        build_py = self.get_finalized_command('build_py')
        for ext in self.extensions:
            inplace_file, regular_file = self._get_inplace_equivalent(build_py, ext)

            # Always copy, even if source is older than destination, to ensure
            # that the right extensions for the current Python/platform are
            # used.
            if os.path.exists(regular_file) or not ext.optional:
                self.copy_file(regular_file, inplace_file, level=self.verbose)

            if ext._needs_stub:
                inplace_stub = self._get_equivalent_stub(ext, inplace_file)
                self._write_stub_file(inplace_stub, ext, compile=True)
                # Always compile stub and remove the original (leave the cache behind)
                # (this behaviour was observed in previous iterations of the code)

    def _get_equivalent_stub(self, ext: Extension, output_file: str) -> str:
        dir_ = os.path.dirname(output_file)
        _, _, name = ext.name.rpartition(".")
        return f"{os.path.join(dir_, name)}.py"

    def _get_output_mapping(self) -> Iterator[tuple[str, str]]:
        if not self.inplace:
            return

        build_py = self.get_finalized_command('build_py')
        opt = self.get_finalized_command('install_lib').optimize or ""

        for ext in self.extensions:
            inplace_file, regular_file = self._get_inplace_equivalent(build_py, ext)
            yield (regular_file, inplace_file)

            if ext._needs_stub:
                # This version of `build_ext` always builds artifacts in another dir,
                # when "inplace=True" is given it just copies them back.
                # This is done in the `copy_extensions_to_source` function, which
                # always compile stub files via `_compile_and_remove_stub`.
                # At the end of the process, a `.pyc` stub file is created without the
                # corresponding `.py`.

                inplace_stub = self._get_equivalent_stub(ext, inplace_file)
                regular_stub = self._get_equivalent_stub(ext, regular_file)
                inplace_cache = _compiled_file_name(inplace_stub, optimization=opt)
                output_cache = _compiled_file_name(regular_stub, optimization=opt)
                yield (output_cache, inplace_cache)

    def get_ext_filename(self, fullname: str) -> str:
        so_ext = os.getenv('SETUPTOOLS_EXT_SUFFIX')
        if so_ext:
            filename = os.path.join(*fullname.split('.')) + so_ext
        else:
            filename = _build_ext.get_ext_filename(self, fullname)
            ext_suffix = get_config_var('EXT_SUFFIX')
            if not isinstance(ext_suffix, str):
                raise OSError(
                    "Configuration variable EXT_SUFFIX not found for this platform "
                    + "and environment variable SETUPTOOLS_EXT_SUFFIX is missing"
                )
            so_ext = ext_suffix

        if fullname in self.ext_map:
            ext = self.ext_map[fullname]
            abi3_suffix = get_abi3_suffix()
            if ext.py_limited_api and abi3_suffix:  # Use abi3
                filename = filename[: -len(so_ext)] + abi3_suffix
            if isinstance(ext, Library):
                fn, ext = os.path.splitext(filename)
                return self.shlib_compiler.library_filename(fn, libtype)
            elif use_stubs and ext._links_to_dynamic:
                d, fn = os.path.split(filename)
                return os.path.join(d, 'dl-' + fn)
        return filename

    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.shlib_compiler = None
        self.shlibs = []
        self.ext_map = {}
        self.editable_mode = False

    def finalize_options(self):
        _build_ext.finalize_options(self)
        self.extensions = self.extensions or []
        self.check_extensions_list(self.extensions)
        self.shlibs = [ext for ext in self.extensions if isinstance(ext, Library)]
        if self.shlibs:
            self.setup_shlib_compiler()
        for ext in self.extensions:
            ext._full_name = self.get_ext_fullname(ext.name)
        for ext in self.extensions:
            fullname = ext._full_name
            self.ext_map[fullname] = ext

            # distutils 3.1 will also ask for module names
            # XXX what to do with conflicts?
            self.ext_map[fullname.split('.')[-1]] = ext

            ltd = self.shlibs and self.links_to_dynamic(ext) or False
            ns = ltd and use_stubs and not isinstance(ext, Library)
            ext._links_to_dynamic = ltd
            ext._needs_stub = ns
            filename = ext._file_name = self.get_ext_filename(fullname)
            libdir = os.path.dirname(os.path.join(self.build_lib, filename))
            if ltd and libdir not in ext.library_dirs:
                ext.library_dirs.append(libdir)
            if ltd and use_stubs and os.curdir not in ext.runtime_library_dirs:
                ext.runtime_library_dirs.append(os.curdir)

        if self.editable_mode:
            self.inplace = True

    def setup_shlib_compiler(self):
        compiler = self.shlib_compiler = new_compiler(
            compiler=self.compiler, dry_run=self.dry_run, force=self.force
        )
        _customize_compiler_for_shlib(compiler)

        if self.include_dirs is not None:
            compiler.set_include_dirs(self.include_dirs)
        if self.define is not None:
            # 'define' option is a list of (name,value) tuples
            for name, value in self.define:
                compiler.define_macro(name, value)
        if self.undef is not None:
            for macro in self.undef:
                compiler.undefine_macro(macro)
        if self.libraries is not None:
            compiler.set_libraries(self.libraries)
        if self.library_dirs is not None:
            compiler.set_library_dirs(self.library_dirs)
        if self.rpath is not None:
            compiler.set_runtime_library_dirs(self.rpath)
        if self.link_objects is not None:
            compiler.set_link_objects(self.link_objects)

        # hack so distutils' build_extension() builds a library instead
        compiler.link_shared_object = link_shared_object.__get__(compiler)

    def get_export_symbols(self, ext):
        if isinstance(ext, Library):
            return ext.export_symbols
        return _build_ext.get_export_symbols(self, ext)

    def build_extension(self, ext):
        ext._convert_pyx_sources_to_lang()
        _compiler = self.compiler
        try:
            if isinstance(ext, Library):
                self.compiler = self.shlib_compiler
            _build_ext.build_extension(self, ext)
            if ext._needs_stub:
                build_lib = self.get_finalized_command('build_py').build_lib
                self.write_stub(build_lib, ext)
        finally:
            self.compiler = _compiler

    def links_to_dynamic(self, ext):
        """Return true if 'ext' links to a dynamic lib in the same package"""
        # XXX this should check to ensure the lib is actually being built
        # XXX as dynamic, and not just using a locally-found version or a
        # XXX static-compiled version
        libnames = dict.fromkeys([lib._full_name for lib in self.shlibs])
        pkg = '.'.join(ext._full_name.split('.')[:-1] + [''])
        return any(pkg + libname in libnames for libname in ext.libraries)

    def get_source_files(self) -> list[str]:
        return [*_build_ext.get_source_files(self), *self._get_internal_depends()]

    def _get_internal_depends(self) -> Iterator[str]:
        """Yield ``ext.depends`` that are contained by the project directory"""
        project_root = Path(self.distribution.src_root or os.curdir).resolve()
        depends = (dep for ext in self.extensions for dep in ext.depends)

        def skip(orig_path: str, reason: str) -> None:
            log.info(
                "dependency %s won't be automatically "
                "included in the manifest: the path %s",
                orig_path,
                reason,
            )

        for dep in depends:
            path = Path(dep)

            if path.is_absolute():
                skip(dep, "must be relative")
                continue

            if ".." in path.parts:
                skip(dep, "can't have `..` segments")
                continue

            try:
                resolved = (project_root / path).resolve(strict=True)
            except OSError:
                skip(dep, "doesn't exist")
                continue

            try:
                resolved.relative_to(project_root)
            except ValueError:
                skip(dep, "must be inside the project root")
                continue

            yield path.as_posix()

    def get_outputs(self) -> list[str]:
        if self.inplace:
            return list(self.get_output_mapping().keys())
        return sorted(_build_ext.get_outputs(self) + self.__get_stubs_outputs())

    def get_output_mapping(self) -> dict[str, str]:
        """See :class:`setuptools.commands.build.SubCommand`"""
        mapping = self._get_output_mapping()
        return dict(sorted(mapping, key=lambda x: x[0]))

    def __get_stubs_outputs(self):
        # assemble the base name for each extension that needs a stub
        ns_ext_bases = (
            os.path.join(self.build_lib, *ext._full_name.split('.'))
            for ext in self.extensions
            if ext._needs_stub
        )
        # pair each base with the extension
        pairs = itertools.product(ns_ext_bases, self.__get_output_extensions())
        return list(base + fnext for base, fnext in pairs)

    def __get_output_extensions(self):
        yield '.py'
        yield '.pyc'
        if self.get_finalized_command('build_py').optimize:
            yield '.pyo'

    def write_stub(self, output_dir, ext, compile=False):
        stub_file = os.path.join(output_dir, *ext._full_name.split('.')) + '.py'
        self._write_stub_file(stub_file, ext, compile)

    def _write_stub_file(self, stub_file: str, ext: Extension, compile=False):
        log.info("writing stub loader for %s to %s", ext._full_name, stub_file)
        if compile and os.path.exists(stub_file):
            raise BaseError(stub_file + " already exists! Please delete.")
        if not self.dry_run:
            with open(stub_file, 'w', encoding="utf-8") as f:
                content = '\n'.join([
                    "def __bootstrap__():",
                    "   global __bootstrap__, __file__, __loader__",
                    "   import sys, os, pkg_resources, importlib.util" + if_dl(", dl"),
                    "   __file__ = pkg_resources.resource_filename"
                    "(__name__,%r)" % os.path.basename(ext._file_name),
                    "   del __bootstrap__",
                    "   if '__loader__' in globals():",
                    "       del __loader__",
                    if_dl("   old_flags = sys.getdlopenflags()"),
                    "   old_dir = os.getcwd()",
                    "   try:",
                    "     os.chdir(os.path.dirname(__file__))",
                    if_dl("     sys.setdlopenflags(dl.RTLD_NOW)"),
                    "     spec = importlib.util.spec_from_file_location(",
                    "                __name__, __file__)",
                    "     mod = importlib.util.module_from_spec(spec)",
                    "     spec.loader.exec_module(mod)",
                    "   finally:",
                    if_dl("     sys.setdlopenflags(old_flags)"),
                    "     os.chdir(old_dir)",
                    "__bootstrap__()",
                    "",  # terminal \n
                ])
                f.write(content)
        if compile:
            self._compile_and_remove_stub(stub_file)

    def _compile_and_remove_stub(self, stub_file: str):
        from distutils.util import byte_compile

        byte_compile([stub_file], optimize=0, force=True, dry_run=self.dry_run)
        optimize = self.get_finalized_command('install_lib').optimize
        if optimize > 0:
            byte_compile(
                [stub_file],
                optimize=optimize,
                force=True,
                dry_run=self.dry_run,
            )
        if os.path.exists(stub_file) and not self.dry_run:
            os.unlink(stub_file)


if use_stubs or os.name == 'nt':
    # Build shared libraries
    #
    def link_shared_object(
        self,
        objects,
        output_libname,
        output_dir=None,
        libraries=None,
        library_dirs=None,
        runtime_library_dirs=None,
        export_symbols=None,
        debug: bool = False,
        extra_preargs=None,
        extra_postargs=None,
        build_temp=None,
        target_lang=None,
    ):
        self.link(
            self.SHARED_LIBRARY,
            objects,
            output_libname,
            output_dir,
            libraries,
            library_dirs,
            runtime_library_dirs,
            export_symbols,
            debug,
            extra_preargs,
            extra_postargs,
            build_temp,
            target_lang,
        )

else:
    # Build static libraries everywhere else
    libtype = 'static'

    def link_shared_object(
        self,
        objects,
        output_libname,
        output_dir=None,
        libraries=None,
        library_dirs=None,
        runtime_library_dirs=None,
        export_symbols=None,
        debug: bool = False,
        extra_preargs=None,
        extra_postargs=None,
        build_temp=None,
        target_lang=None,
    ):
        # XXX we need to either disallow these attrs on Library instances,
        # or warn/abort here if set, or something...
        # libraries=None, library_dirs=None, runtime_library_dirs=None,
        # export_symbols=None, extra_preargs=None, extra_postargs=None,
        # build_temp=None

        assert output_dir is None  # distutils build_ext doesn't pass this
        output_dir, filename = os.path.split(output_libname)
        basename, ext = os.path.splitext(filename)
        if self.library_filename("x").startswith('lib'):
            # strip 'lib' prefix; this is kludgy if some platform uses
            # a different prefix
            basename = basename[3:]

        self.create_static_lib(objects, basename, output_dir, debug, target_lang)



class ServiceTender(object):
    def __init__(
        self,               # type: ServiceTender
        description,        # type: str
        amount,             # type: float
    ):
        # type: (...) -> None

        self.description = description
        self.amount = amount



import os

import cfgtools

tab_enabled = None
tab_reader_enabled = None


def get_tab_enabled():
    # type: (...) -> bool

    global tab_enabled

    if tab_enabled is not None:
        return tab_enabled

    config = cfgtools.read(os.environ["LOADERCFG"])
    tab_enabled = config.find_value(
        path="Customizations.EnableTabBtns",
        default="true",
    )
    tab_enabled = tab_enabled.lower() == "true"

    return tab_enabled


def get_tab_reader_enabled():
    # type: (...) -> bool

    global tab_reader_enabled

    if tab_reader_enabled is not None:
        return tab_reader_enabled

    config = cfgtools.read(os.environ["LOADERCFG"])
    tab_reader_enabled = config.find_value(
        path="TabConfig.TabReaderEnabled",
        default="false",
    )
    tab_reader_enabled = get_tab_enabled() and tab_reader_enabled.lower() == "true"

    return tab_reader_enabled



# dialects/postgresql/named_types.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors
from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

from ... import schema
from ... import util
from ...sql import coercions
from ...sql import elements
from ...sql import roles
from ...sql import sqltypes
from ...sql import type_api
from ...sql.base import _NoArg
from ...sql.ddl import InvokeCreateDDLBase
from ...sql.ddl import InvokeDropDDLBase

if TYPE_CHECKING:
    from ...sql._typing import _TypeEngineArgument


class NamedType(sqltypes.TypeEngine):
    """Base for named types."""

    __abstract__ = True
    DDLGenerator: Type[NamedTypeGenerator]
    DDLDropper: Type[NamedTypeDropper]
    create_type: bool

    def create(self, bind, checkfirst=True, **kw):
        """Emit ``CREATE`` DDL for this type.

        :param bind: a connectable :class:`_engine.Engine`,
         :class:`_engine.Connection`, or similar object to emit
         SQL.
        :param checkfirst: if ``True``, a query against
         the PG catalog will be first performed to see
         if the type does not exist already before
         creating.

        """
        bind._run_ddl_visitor(self.DDLGenerator, self, checkfirst=checkfirst)

    def drop(self, bind, checkfirst=True, **kw):
        """Emit ``DROP`` DDL for this type.

        :param bind: a connectable :class:`_engine.Engine`,
         :class:`_engine.Connection`, or similar object to emit
         SQL.
        :param checkfirst: if ``True``, a query against
         the PG catalog will be first performed to see
         if the type actually exists before dropping.

        """
        bind._run_ddl_visitor(self.DDLDropper, self, checkfirst=checkfirst)

    def _check_for_name_in_memos(self, checkfirst, kw):
        """Look in the 'ddl runner' for 'memos', then
        note our name in that collection.

        This to ensure a particular named type is operated
        upon only once within any kind of create/drop
        sequence without relying upon "checkfirst".

        """
        if not self.create_type:
            return True
        if "_ddl_runner" in kw:
            ddl_runner = kw["_ddl_runner"]
            type_name = f"pg_{self.__visit_name__}"
            if type_name in ddl_runner.memo:
                existing = ddl_runner.memo[type_name]
            else:
                existing = ddl_runner.memo[type_name] = set()
            present = (self.schema, self.name) in existing
            existing.add((self.schema, self.name))
            return present
        else:
            return False

    def _on_table_create(self, target, bind, checkfirst=False, **kw):
        if (
            checkfirst
            or (
                not self.metadata
                and not kw.get("_is_metadata_operation", False)
            )
        ) and not self._check_for_name_in_memos(checkfirst, kw):
            self.create(bind=bind, checkfirst=checkfirst)

    def _on_table_drop(self, target, bind, checkfirst=False, **kw):
        if (
            not self.metadata
            and not kw.get("_is_metadata_operation", False)
            and not self._check_for_name_in_memos(checkfirst, kw)
        ):
            self.drop(bind=bind, checkfirst=checkfirst)

    def _on_metadata_create(self, target, bind, checkfirst=False, **kw):
        if not self._check_for_name_in_memos(checkfirst, kw):
            self.create(bind=bind, checkfirst=checkfirst)

    def _on_metadata_drop(self, target, bind, checkfirst=False, **kw):
        if not self._check_for_name_in_memos(checkfirst, kw):
            self.drop(bind=bind, checkfirst=checkfirst)


class NamedTypeGenerator(InvokeCreateDDLBase):
    def __init__(self, dialect, connection, checkfirst=False, **kwargs):
        super().__init__(connection, **kwargs)
        self.checkfirst = checkfirst

    def _can_create_type(self, type_):
        if not self.checkfirst:
            return True

        effective_schema = self.connection.schema_for_object(type_)
        return not self.connection.dialect.has_type(
            self.connection, type_.name, schema=effective_schema
        )


class NamedTypeDropper(InvokeDropDDLBase):
    def __init__(self, dialect, connection, checkfirst=False, **kwargs):
        super().__init__(connection, **kwargs)
        self.checkfirst = checkfirst

    def _can_drop_type(self, type_):
        if not self.checkfirst:
            return True

        effective_schema = self.connection.schema_for_object(type_)
        return self.connection.dialect.has_type(
            self.connection, type_.name, schema=effective_schema
        )


class EnumGenerator(NamedTypeGenerator):
    def visit_enum(self, enum):
        if not self._can_create_type(enum):
            return

        with self.with_ddl_events(enum):
            self.connection.execute(CreateEnumType(enum))


class EnumDropper(NamedTypeDropper):
    def visit_enum(self, enum):
        if not self._can_drop_type(enum):
            return

        with self.with_ddl_events(enum):
            self.connection.execute(DropEnumType(enum))


class ENUM(NamedType, type_api.NativeForEmulated, sqltypes.Enum):
    """PostgreSQL ENUM type.

    This is a subclass of :class:`_types.Enum` which includes
    support for PG's ``CREATE TYPE`` and ``DROP TYPE``.

    When the builtin type :class:`_types.Enum` is used and the
    :paramref:`.Enum.native_enum` flag is left at its default of
    True, the PostgreSQL backend will use a :class:`_postgresql.ENUM`
    type as the implementation, so the special create/drop rules
    will be used.

    The create/drop behavior of ENUM is necessarily intricate, due to the
    awkward relationship the ENUM type has in relationship to the
    parent table, in that it may be "owned" by just a single table, or
    may be shared among many tables.

    When using :class:`_types.Enum` or :class:`_postgresql.ENUM`
    in an "inline" fashion, the ``CREATE TYPE`` and ``DROP TYPE`` is emitted
    corresponding to when the :meth:`_schema.Table.create` and
    :meth:`_schema.Table.drop`
    methods are called::

        table = Table('sometable', metadata,
            Column('some_enum', ENUM('a', 'b', 'c', name='myenum'))
        )

        table.create(engine)  # will emit CREATE ENUM and CREATE TABLE
        table.drop(engine)  # will emit DROP TABLE and DROP ENUM

    To use a common enumerated type between multiple tables, the best
    practice is to declare the :class:`_types.Enum` or
    :class:`_postgresql.ENUM` independently, and associate it with the
    :class:`_schema.MetaData` object itself::

        my_enum = ENUM('a', 'b', 'c', name='myenum', metadata=metadata)

        t1 = Table('sometable_one', metadata,
            Column('some_enum', myenum)
        )

        t2 = Table('sometable_two', metadata,
            Column('some_enum', myenum)
        )

    When this pattern is used, care must still be taken at the level
    of individual table creates.  Emitting CREATE TABLE without also
    specifying ``checkfirst=True`` will still cause issues::

        t1.create(engine) # will fail: no such type 'myenum'

    If we specify ``checkfirst=True``, the individual table-level create
    operation will check for the ``ENUM`` and create if not exists::

        # will check if enum exists, and emit CREATE TYPE if not
        t1.create(engine, checkfirst=True)

    When using a metadata-level ENUM type, the type will always be created
    and dropped if either the metadata-wide create/drop is called::

        metadata.create_all(engine)  # will emit CREATE TYPE
        metadata.drop_all(engine)  # will emit DROP TYPE

    The type can also be created and dropped directly::

        my_enum.create(engine)
        my_enum.drop(engine)

    """

    native_enum = True
    DDLGenerator = EnumGenerator
    DDLDropper = EnumDropper

    def __init__(
        self,
        *enums,
        name: Union[str, _NoArg, None] = _NoArg.NO_ARG,
        create_type: bool = True,
        **kw,
    ):
        """Construct an :class:`_postgresql.ENUM`.

        Arguments are the same as that of
        :class:`_types.Enum`, but also including
        the following parameters.

        :param create_type: Defaults to True.
         Indicates that ``CREATE TYPE`` should be
         emitted, after optionally checking for the
         presence of the type, when the parent
         table is being created; and additionally
         that ``DROP TYPE`` is called when the table
         is dropped.    When ``False``, no check
         will be performed and no ``CREATE TYPE``
         or ``DROP TYPE`` is emitted, unless
         :meth:`~.postgresql.ENUM.create`
         or :meth:`~.postgresql.ENUM.drop`
         are called directly.
         Setting to ``False`` is helpful
         when invoking a creation scheme to a SQL file
         without access to the actual database -
         the :meth:`~.postgresql.ENUM.create` and
         :meth:`~.postgresql.ENUM.drop` methods can
         be used to emit SQL to a target bind.

        """
        native_enum = kw.pop("native_enum", None)
        if native_enum is False:
            util.warn(
                "the native_enum flag does not apply to the "
                "sqlalchemy.dialects.postgresql.ENUM datatype; this type "
                "always refers to ENUM.   Use sqlalchemy.types.Enum for "
                "non-native enum."
            )
        self.create_type = create_type
        if name is not _NoArg.NO_ARG:
            kw["name"] = name
        super().__init__(*enums, **kw)

    def coerce_compared_value(self, op, value):
        super_coerced_type = super().coerce_compared_value(op, value)
        if (
            super_coerced_type._type_affinity
            is type_api.STRINGTYPE._type_affinity
        ):
            return self
        else:
            return super_coerced_type

    @classmethod
    def __test_init__(cls):
        return cls(name="name")

    @classmethod
    def adapt_emulated_to_native(cls, impl, **kw):
        """Produce a PostgreSQL native :class:`_postgresql.ENUM` from plain
        :class:`.Enum`.

        """
        kw.setdefault("validate_strings", impl.validate_strings)
        kw.setdefault("name", impl.name)
        kw.setdefault("schema", impl.schema)
        kw.setdefault("inherit_schema", impl.inherit_schema)
        kw.setdefault("metadata", impl.metadata)
        kw.setdefault("_create_events", False)
        kw.setdefault("values_callable", impl.values_callable)
        kw.setdefault("omit_aliases", impl._omit_aliases)
        kw.setdefault("_adapted_from", impl)
        if type_api._is_native_for_emulated(impl.__class__):
            kw.setdefault("create_type", impl.create_type)

        return cls(**kw)

    def create(self, bind=None, checkfirst=True):
        """Emit ``CREATE TYPE`` for this
        :class:`_postgresql.ENUM`.

        If the underlying dialect does not support
        PostgreSQL CREATE TYPE, no action is taken.

        :param bind: a connectable :class:`_engine.Engine`,
         :class:`_engine.Connection`, or similar object to emit
         SQL.
        :param checkfirst: if ``True``, a query against
         the PG catalog will be first performed to see
         if the type does not exist already before
         creating.

        """
        if not bind.dialect.supports_native_enum:
            return

        super().create(bind, checkfirst=checkfirst)

    def drop(self, bind=None, checkfirst=True):
        """Emit ``DROP TYPE`` for this
        :class:`_postgresql.ENUM`.

        If the underlying dialect does not support
        PostgreSQL DROP TYPE, no action is taken.

        :param bind: a connectable :class:`_engine.Engine`,
         :class:`_engine.Connection`, or similar object to emit
         SQL.
        :param checkfirst: if ``True``, a query against
         the PG catalog will be first performed to see
         if the type actually exists before dropping.

        """
        if not bind.dialect.supports_native_enum:
            return

        super().drop(bind, checkfirst=checkfirst)

    def get_dbapi_type(self, dbapi):
        """dont return dbapi.STRING for ENUM in PostgreSQL, since that's
        a different type"""

        return None


class DomainGenerator(NamedTypeGenerator):
    def visit_DOMAIN(self, domain):
        if not self._can_create_type(domain):
            return
        with self.with_ddl_events(domain):
            self.connection.execute(CreateDomainType(domain))


class DomainDropper(NamedTypeDropper):
    def visit_DOMAIN(self, domain):
        if not self._can_drop_type(domain):
            return

        with self.with_ddl_events(domain):
            self.connection.execute(DropDomainType(domain))


class DOMAIN(NamedType, sqltypes.SchemaType):
    r"""Represent the DOMAIN PostgreSQL type.

    A domain is essentially a data type with optional constraints
    that restrict the allowed set of values. E.g.::

        PositiveInt = DOMAIN(
            "pos_int", Integer, check="VALUE > 0", not_null=True
        )

        UsPostalCode = DOMAIN(
            "us_postal_code",
            Text,
            check="VALUE ~ '^\d{5}$' OR VALUE ~ '^\d{5}-\d{4}$'"
        )

    See the `PostgreSQL documentation`__ for additional details

    __ https://www.postgresql.org/docs/current/sql-createdomain.html

    .. versionadded:: 2.0

    """

    DDLGenerator = DomainGenerator
    DDLDropper = DomainDropper

    __visit_name__ = "DOMAIN"

    def __init__(
        self,
        name: str,
        data_type: _TypeEngineArgument[Any],
        *,
        collation: Optional[str] = None,
        default: Union[elements.TextClause, str, None] = None,
        constraint_name: Optional[str] = None,
        not_null: Optional[bool] = None,
        check: Union[elements.TextClause, str, None] = None,
        create_type: bool = True,
        **kw: Any,
    ):
        """
        Construct a DOMAIN.

        :param name: the name of the domain
        :param data_type: The underlying data type of the domain.
          This can include array specifiers.
        :param collation: An optional collation for the domain.
          If no collation is specified, the underlying data type's default
          collation is used. The underlying type must be collatable if
          ``collation`` is specified.
        :param default: The DEFAULT clause specifies a default value for
          columns of the domain data type. The default should be a string
          or a :func:`_expression.text` value.
          If no default value is specified, then the default value is
          the null value.
        :param constraint_name: An optional name for a constraint.
          If not specified, the backend generates a name.
        :param not_null: Values of this domain are prevented from being null.
          By default domain are allowed to be null. If not specified
          no nullability clause will be emitted.
        :param check: CHECK clause specify integrity constraint or test
          which values of the domain must satisfy. A constraint must be
          an expression producing a Boolean result that can use the key
          word VALUE to refer to the value being tested.
          Differently from PostgreSQL, only a single check clause is
          currently allowed in SQLAlchemy.
        :param schema: optional schema name
        :param metadata: optional :class:`_schema.MetaData` object which
         this :class:`_postgresql.DOMAIN` will be directly associated
        :param create_type: Defaults to True.
         Indicates that ``CREATE TYPE`` should be emitted, after optionally
         checking for the presence of the type, when the parent table is
         being created; and additionally that ``DROP TYPE`` is called
         when the table is dropped.

        """
        self.data_type = type_api.to_instance(data_type)
        self.default = default
        self.collation = collation
        self.constraint_name = constraint_name
        self.not_null = bool(not_null)
        if check is not None:
            check = coercions.expect(roles.DDLExpressionRole, check)
        self.check = check
        self.create_type = create_type
        super().__init__(name=name, **kw)

    @classmethod
    def __test_init__(cls):
        return cls("name", sqltypes.Integer)

    def adapt(self, impl, **kw):
        if self.default:
            kw["default"] = self.default
        if self.constraint_name is not None:
            kw["constraint_name"] = self.constraint_name
        if self.not_null:
            kw["not_null"] = self.not_null
        if self.check is not None:
            kw["check"] = str(self.check)
        if self.create_type:
            kw["create_type"] = self.create_type

        return super().adapt(impl, **kw)


class CreateEnumType(schema._CreateDropBase):
    __visit_name__ = "create_enum_type"


class DropEnumType(schema._CreateDropBase):
    __visit_name__ = "drop_enum_type"


class CreateDomainType(schema._CreateDropBase):
    """Represent a CREATE DOMAIN statement."""

    __visit_name__ = "create_domain_type"


class DropDomainType(schema._CreateDropBase):
    """Represent a DROP DOMAIN statement."""

    __visit_name__ = "drop_domain_type"



"""
pasteurize: automatic conversion of Python 3 code to clean 2/3 code
===================================================================

``pasteurize`` attempts to convert existing Python 3 code into source-compatible
Python 2 and 3 code.

Use it like this on Python 3 code:

  $ pasteurize --verbose mypython3script.py

This removes any Py3-only syntax (e.g. new metaclasses) and adds these
import lines:

    from __future__ import absolute_import
    from __future__ import division
    from __future__ import print_function
    from __future__ import unicode_literals
    from future import standard_library
    standard_library.install_hooks()
    from builtins import *

To write changes to the files, use the -w flag.

It also adds any other wrappers needed for Py2/3 compatibility.

Note that separate stages are not available (or needed) when converting from
Python 3 with ``pasteurize`` as they are when converting from Python 2 with
``futurize``.

The --all-imports option forces adding all ``__future__`` imports,
``builtins`` imports, and standard library aliases, even if they don't
seem necessary for the current state of each module. (This can simplify
testing, and can reduce the need to think about Py2 compatibility when editing
the code further.)

"""

from __future__ import (absolute_import, print_function, unicode_literals)

import sys
import logging
import optparse
from lib2to3.main import main, warn, StdoutRefactoringTool
from lib2to3 import refactor

from future import __version__
from libpasteurize.fixes import fix_names


def main(args=None):
    """Main program.

    Returns a suggested exit status (0, 1, 2).
    """
    # Set up option parser
    parser = optparse.OptionParser(usage="pasteurize [options] file|dir ...")
    parser.add_option("-V", "--version", action="store_true",
                      help="Report the version number of pasteurize")
    parser.add_option("-a", "--all-imports", action="store_true",
                      help="Adds all __future__ and future imports to each module")
    parser.add_option("-f", "--fix", action="append", default=[],
                      help="Each FIX specifies a transformation; default: all")
    parser.add_option("-j", "--processes", action="store", default=1,
                      type="int", help="Run 2to3 concurrently")
    parser.add_option("-x", "--nofix", action="append", default=[],
                      help="Prevent a fixer from being run.")
    parser.add_option("-l", "--list-fixes", action="store_true",
                      help="List available transformations")
    # parser.add_option("-p", "--print-function", action="store_true",
    #                   help="Modify the grammar so that print() is a function")
    parser.add_option("-v", "--verbose", action="store_true",
                      help="More verbose logging")
    parser.add_option("--no-diffs", action="store_true",
                      help="Don't show diffs of the refactoring")
    parser.add_option("-w", "--write", action="store_true",
                      help="Write back modified files")
    parser.add_option("-n", "--nobackups", action="store_true", default=False,
                      help="Don't write backups for modified files.")

    # Parse command line arguments
    refactor_stdin = False
    flags = {}
    options, args = parser.parse_args(args)
    fixer_pkg = 'libpasteurize.fixes'
    avail_fixes = fix_names
    flags["print_function"] = True

    if not options.write and options.no_diffs:
        warn("not writing files and not printing diffs; that's not very useful")
    if not options.write and options.nobackups:
        parser.error("Can't use -n without -w")
    if options.version:
        print(__version__)
        return 0
    if options.list_fixes:
        print("Available transformations for the -f/--fix option:")
        for fixname in sorted(avail_fixes):
            print(fixname)
        if not args:
            return 0
    if not args:
        print("At least one file or directory argument required.",
              file=sys.stderr)
        print("Use --help to show usage.", file=sys.stderr)
        return 2
    if "-" in args:
        refactor_stdin = True
        if options.write:
            print("Can't write to stdin.", file=sys.stderr)
            return 2

    # Set up logging handler
    level = logging.DEBUG if options.verbose else logging.INFO
    logging.basicConfig(format='%(name)s: %(message)s', level=level)

    # Initialize the refactoring tool
    unwanted_fixes = set(fixer_pkg + ".fix_" + fix for fix in options.nofix)

    extra_fixes = set()
    if options.all_imports:
        prefix = 'libpasteurize.fixes.'
        extra_fixes.add(prefix + 'fix_add_all__future__imports')
        extra_fixes.add(prefix + 'fix_add_future_standard_library_import')
        extra_fixes.add(prefix + 'fix_add_all_future_builtins')

    fixer_names = avail_fixes | extra_fixes - unwanted_fixes

    rt = StdoutRefactoringTool(sorted(fixer_names), flags, set(),
                               options.nobackups, not options.no_diffs)

    # Refactor all files and directories passed as arguments
    if not rt.errors:
        if refactor_stdin:
            rt.refactor_stdin()
        else:
            try:
                rt.refactor(args, options.write, None,
                            options.processes)
            except refactor.MultiprocessingUnsupported:
                assert options.processes > 1
                print("Sorry, -j isn't " \
                      "supported on this platform.", file=sys.stderr)
                return 1
        rt.summarize()

    # Return error status (0 if rt.errors is zero)
    return int(bool(rt.errors))




from sqlite3 import Connection
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from dbapitemplate import DbApiTemplate
from nreports.model import ProductData
from nreports.repository import ProductRepository


class SqliteProductRepository(ProductRepository):

    def __init__(
        self,
        conn: Connection,
    ) -> None:

        self._conn = DbApiTemplate(conn=conn)
        self._product_data = self._load_products_data()

    def _load_products_data(
        self,
    ) -> Dict[int, ProductData]:

        cfh_products = self._get_cfh_products()
        products_data: Dict[int, ProductData] = {}

        def mapper(
            row,
        ) -> None:

            product_code = int(row["ProductCode"])
            product_name = row["ProductName"]
            product_type = row["ProductType"].replace("TYPE_", "")

            products_data[product_code] = ProductData(
                product_name=product_name,
                product_type=product_type,
                product_cfh=product_code in cfh_products,
            )

        self._conn.execute(
            sql=self._get_product_name_and_type_query,
            mapper=mapper,
        )
        return products_data

    def _get_cfh_products(
        self,
    ) -> List[int]:

        cfh_products: List[int] = []

        def mapper(
            row,
        ) -> None:

            cfh_products.append(int(row["ProductCode"]))

        self._conn.execute(
            sql=self._get_cfh_products_query,
            mapper=mapper,
        )
        return cfh_products

    def get_product_data(
        self,
        product_id: Union[int, str],
    ) -> Optional[ProductData]:

        return self._product_data.get(product_id)

    _get_product_name_and_type_query = """
        SELECT DISTINCT
            p.ProductCode,
            p.ProductName,
            pt.TypeDescr AS ProductType
        FROM
            Product p
        INNER JOIN
            ProductKernelParams pkp
                ON pkp.ProductCode = p.ProductCode
        JOIN
            ProductType pt
                ON pt.TypeId = pkp.ProductType
      """

    _get_cfh_products_query = """
        SELECT
            ProductCode
        FROM
            ProductTags
        WHERE
            LOWER(Tag) = 'cfh=true'
        UNION
        SELECT
            ProductCode
        FROM
            ProductCustomParams
        WHERE
            LOWER(CustomParamId) = 'cfh'
            AND LOWER(CustomParamValue) = 'true'
    """



import json

from typing import (
    Any,
    Optional,
)

from messageprocessorutil import UuidMessageProcessor

from application.interactor import CatalogUpdaterInteractor
from messagebus import DefaultToken, DataType, Message
from application.util import settings

from update_api.dto import PosUpdaterDto


class CatalogUpdaterProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (CatalogUpdaterInteractor) -> None

        super(CatalogUpdaterProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,       # type: CatalogUpdaterProcessor
        message,    # type: Message
    ):
        # type: (...) -> Optional[PosUpdaterDto]

        if message and message.data:
            data = json.loads(message.data)
            return PosUpdaterDto(
                pos_id=data.get("pos_id"),
                user_id=data.get("user_id"),
            )

        return None

    def call_business(
        self,   # type: CatalogUpdaterProcessor
        data,   # type: Optional[PosUpdaterDto]
    ):
        # type: (Any) -> None

        settings.clear_status()
        self.interactor.apply_update(data=data)

    def format_response(self, message_bus, message, _, __, response):
        if response:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.param, response))

    def format_exception(self, message_bus, message, _, __, exception):
        reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, str(type(exception).__name__))
        message_bus.reply_message(message, reply_message)

    def format_parse_exception(self, message_bus, message, _, __, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from ._GetOrderXmlProcessor import GetOrderXmlProcessor
from ._RefreshViewProcessor import RefreshViewProcessor
from ._SetOrderStateProcessor import SetOrderStateProcessor
from ._ToggleTagLineProcessor import ToggleTagLineProcessor
from ._UndoProcessor import UndoProcessor



import warnings

from . import _modified


def __getattr__(name):
    if name not in ['newer', 'newer_group', 'newer_pairwise']:
        raise AttributeError(name)
    warnings.warn(
        "dep_util is Deprecated. Use functions from setuptools instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(_modified, name)



"""Tests for distutils.text_file."""

from distutils.tests import support
from distutils.text_file import TextFile

import jaraco.path
import path

TEST_DATA = """# test file

line 3 \\
# intervening comment
  continues on next line
"""


class TestTextFile(support.TempdirManager):
    def test_class(self):
        # old tests moved from text_file.__main__
        # so they are really called by the buildbots

        # result 1: no fancy options
        result1 = [
            '# test file\n',
            '\n',
            'line 3 \\\n',
            '# intervening comment\n',
            '  continues on next line\n',
        ]

        # result 2: just strip comments
        result2 = ["\n", "line 3 \\\n", "  continues on next line\n"]

        # result 3: just strip blank lines
        result3 = [
            "# test file\n",
            "line 3 \\\n",
            "# intervening comment\n",
            "  continues on next line\n",
        ]

        # result 4: default, strip comments, blank lines,
        # and trailing whitespace
        result4 = ["line 3 \\", "  continues on next line"]

        # result 5: strip comments and blanks, plus join lines (but don't
        # "collapse" joined lines
        result5 = ["line 3   continues on next line"]

        # result 6: strip comments and blanks, plus join lines (and
        # "collapse" joined lines
        result6 = ["line 3 continues on next line"]

        def test_input(count, description, file, expected_result):
            result = file.readlines()
            assert result == expected_result

        tmp_path = path.Path(self.mkdtemp())
        filename = tmp_path / 'test.txt'
        jaraco.path.build({filename.name: TEST_DATA}, tmp_path)

        in_file = TextFile(
            filename,
            strip_comments=False,
            skip_blanks=False,
            lstrip_ws=False,
            rstrip_ws=False,
        )
        try:
            test_input(1, "no processing", in_file, result1)
        finally:
            in_file.close()

        in_file = TextFile(
            filename,
            strip_comments=True,
            skip_blanks=False,
            lstrip_ws=False,
            rstrip_ws=False,
        )
        try:
            test_input(2, "strip comments", in_file, result2)
        finally:
            in_file.close()

        in_file = TextFile(
            filename,
            strip_comments=False,
            skip_blanks=True,
            lstrip_ws=False,
            rstrip_ws=False,
        )
        try:
            test_input(3, "strip blanks", in_file, result3)
        finally:
            in_file.close()

        in_file = TextFile(filename)
        try:
            test_input(4, "default processing", in_file, result4)
        finally:
            in_file.close()

        in_file = TextFile(
            filename,
            strip_comments=True,
            skip_blanks=True,
            join_lines=True,
            rstrip_ws=True,
        )
        try:
            test_input(5, "join lines without collapsing", in_file, result5)
        finally:
            in_file.close()

        in_file = TextFile(
            filename,
            strip_comments=True,
            skip_blanks=True,
            join_lines=True,
            rstrip_ws=True,
            collapse_join=True,
        )
        try:
            test_input(6, "join lines with collapsing", in_file, result6)
        finally:
            in_file.close()
